# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:38
# @Author  : sudoskys
# @File    : app.py
# @Software: PyCharm
from dotenv import load_dotenv
from loguru import logger

load_dotenv()

__area__ = "receiver"


def run():
    import asyncio
    from .aps import aps_start
    from .function import FunctionReceiver
    from llmkira.sdk.cache import global_cache_runtime, global_mongodb_runtime
    global_cache_runtime.init_cache(verbose=True)
    global_mongodb_runtime.init_mongodb(verbose=True)
    func = [
        aps_start(),
        FunctionReceiver().function(),
    ]

    from llmkira.sdk.func_calling import load_plugins, load_from_entrypoint, get_entrypoint_plugins

    from llmkira.setting import StartSetting
    start_setting = StartSetting.from_subdir()
    if start_setting.telegram:
        from .telegram import TelegramReceiver
        func.append(TelegramReceiver().telegram())
    if start_setting.discord:
        from .discord import DiscordReceiver
        func.append(DiscordReceiver().discord())
    if start_setting.kook:
        from .kook import KookReceiver
        func.append(KookReceiver().kook())
    if start_setting.slack:
        from .slack import SlackReceiver
        func.append(SlackReceiver().slack())

    async def _main(_func):
        await asyncio.gather(
            *_func
        )

    # 导入插件
    load_plugins("llmkira/extra/plugins")
    load_from_entrypoint("llmkira.extra.plugin")

    loaded_message = "\n >>".join(get_entrypoint_plugins())
    logger.success(f"\n===========Third Party Plugins Loaded==========\n >>{loaded_message}")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(_main(func))



# -*- coding: utf-8 -*-
# @Time    : 2023/10/21 下午2:14
# @Author  : sudoskys
# @File    : slack.py
# @Software: PyCharm
from typing import Optional

from dotenv import load_dotenv
from loguru import logger
from pydantic import Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class SlackBot(BaseSettings):
    """
    代理设置
    """
    app_token: Optional[str] = Field(None, validation_alias='SLACK_APP_TOKEN')
    # https://api.slack.com/apps

    bot_token: Optional[str] = Field(None, validation_alias='SLACK_BOT_TOKEN')
    # https://api.slack.com/apps/XXXX/oauth?

    secret: Optional[str] = Field(None, validation_alias='SLACK_SIGNING_SECRET')
    # https://api.slack.com/authentication/verifying-requests-from-slack#signing_secrets_admin_page

    proxy_address: Optional[str] = Field(None, validation_alias="SLACK_BOT_PROXY_ADDRESS")  # "all://127.0.0.1:7890"
    bot_id: Optional[str] = Field(None)
    bot_username: Optional[str] = Field(None)
    model_config = SettingsConfigDict(env_file='.env', env_file_encoding='utf-8', extra="ignore")

    @model_validator(mode='after')
    def bot_setting_validator(self):
        try:
            if self.app_token is None:
                raise ValueError("🍀SlackBot `app_token` Not Set")
            if self.bot_token is None:
                raise LookupError("\n🍀SlackBot `bot_token` is empty")
            if self.secret is None:
                raise LookupError("\n🍀SlackBot `secret` is empty")
        except ValueError as e:
            logger.info(str(e))
        except LookupError as e:
            logger.warning(str(e))
        return self

    @property
    def available(self):
        return all([self.app_token, self.bot_token, self.secret])


load_dotenv()
BotSetting = SlackBot()



# -*- coding: utf-8 -*-
from pydantic import ConfigDict

__package__name__ = "llmkira.extra.plugins.translate_file"
__plugin_name__ = "translate_file"
__openapi_version__ = "20231027"

import re

from llmkira.sdk.func_calling import verify_openapi_version

verify_openapi_version(__package__name__, __openapi_version__)
import asyncio
from io import BytesIO
from typing import List

from loguru import logger
from pydantic import BaseModel

from llmkira.schema import RawMessage

from llmkira.sdk.func_calling import BaseTool, PluginMetadata
from llmkira.sdk.func_calling.schema import FuncPair
from llmkira.sdk.schema import File, Function
from llmkira.task import Task, TaskHeader
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from llmkira.sdk.schema import TaskBatch

translate = Function(name=__plugin_name__, description="Help user translate [ReadableFile],only support txt/md")
translate.add_property(
    property_name="language",
    property_description="What language should the text be translated into?",
    property_type="string",
    required=True
)
translate.add_property(
    property_name="file_id",
    property_description="regex: file_id=([a-z0-9]{8}), require ReadableFile[...]",
    property_type="string",
    required=True
)


class Translate(BaseModel):
    language: str
    file_id: str
    model_config = ConfigDict(extra="allow")


class TranslateTool(BaseTool):
    """
    搜索工具
    """
    function: Function = translate
    keywords: list = ["translate", "转换", "convert", "翻译", "译", "md", 'txt']
    file_match_required = re.compile(r".md|.txt")

    def pre_check(self):
        try:
            import nltk
            nltk.download('punkt')
            nltk.download('averaged_perceptron_tagger')
            from unstructured.partition.auto import partition
        except Exception as e:
            logger.error(f"plugin:translate_doc:{e},pls check !pip install unstructured")
            return False
        return True

    def func_message(self, message_text, **kwargs):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self,
                     task: "TaskHeader", receiver: "TaskHeader.Location",
                     exception,
                     env: dict,
                     arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                     **kwargs
                     ):
        try:
            _meta = task.task_meta.reply_notify(
                plugin_name=__plugin_name__,
                callback=TaskHeader.Meta.Callback.create(
                    name=__plugin_name__,
                    function_response=f"Run Failed {exception}",
                    tool_call_id=pending_task.get_batch_id()
                ),
                write_back=True,
                release_chain=True
            )
            await Task(queue=receiver.platform).send_task(
                task=TaskHeader(
                    sender=task.sender,
                    receiver=receiver,
                    task_meta=_meta,
                    message=[
                        RawMessage(
                            user_id=receiver.user_id,
                            chat_id=receiver.chat_id,
                            text=f"🍖{__plugin_name__} Run Failed：{exception}"
                        )
                    ]
                )
            )
        except Exception as e:
            logger.error(e)

    async def translate_docs(self, task: TaskHeader, file: File.Data, target_lang: str):
        if not file.file_name.endswith(('md', "txt")):
            raise ValueError("That Type File is Not supported :-(")

        from unstructured.partition.auto import partition
        elements = partition(file=BytesIO(initial_bytes=file.file_data), include_page_breaks=True)
        write_out_name = f"translated_{file.file_name}"

        write_out_file = BytesIO()
        write_out_file.name = write_out_name
        write_out_list = []

        # FIXME
        llm_task = ...

        async def _fill_box(text):
            try:
                await asyncio.sleep(2)
                result = await llm_task(
                    plugin_name=__plugin_name__,
                    task=task,
                    task_desc=f"Translate text to {target_lang},as origin format",
                    raw_data=text
                )
            except Exception as e:
                logger.error(e)
                result = str(element)
            write_out_list.append(result)

        _buffer = []
        for element in elements:
            _buffer.append(str(element))
            if len("/n".join(_buffer)) >= 1000:
                await _fill_box(text="\n".join(_buffer))
                _buffer = []
        if _buffer:
            await _fill_box(text="\n".join(_buffer))
            _buffer = []

        write_out_file.write("\n\n".join(write_out_list).encode("utf-8"))
        write_out_file.seek(0)
        return write_out_file

    async def callback(self,
                       task: "TaskHeader", receiver: "TaskHeader.Location",
                       env: dict,
                       arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                       **kwargs
                       ):
        return None

    async def run(self,
                  task: "TaskHeader", receiver: "TaskHeader.Location",
                  arg: dict, env: dict, pending_task: "TaskBatch", refer_llm_result: dict = None
                  ):
        """
        处理message，返回message
        """

        _translate_file = []
        for item in task.message:
            if item.file:
                for i in item.file:
                    _translate_file.append(i)
        try:
            translate_arg = Translate.model_validate(arg)
        except Exception:
            raise ValueError("Please specify the following parameters clearly\n file_id=xxx,language=xxx")
        _file_obj = [await i.raw_file()
                     for i in sorted(set(_translate_file), key=_translate_file.index)]
        _file_obj: List[File.Data] = [item for item in _file_obj if item]

        # 处理文件
        _result: List[File] = []
        if not _file_obj:
            return None
        for item in _file_obj:
            translated_file = await self.translate_docs(task=task, file=item, target_lang=translate_arg.language)
            file_obj = await File.upload_file(file_name=translated_file.name,
                                              file_data=translated_file.getvalue(),
                                              creator_uid=receiver.uid
                                              )
            _result.append(file_obj)
        # META
        _meta = task.task_meta.reply_message(
            plugin_name=__plugin_name__,
            callback=TaskHeader.Meta.Callback.create(
                function_response="Translate Success",
                tool_call_id=pending_task.get_batch_id(),
                name=__plugin_name__
            )
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,
                receiver=receiver,
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        file=_result,
                        text="🍖 操作成功！"
                    )
                ]
            )
        )

        logger.debug("Plugin:translate_doc say: {}".format(translate_arg))


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Translate readable file to target language",
    usage=str(TranslateTool().keywords),
    openapi_version=__openapi_version__,
    function={
        FuncPair(function=translate, tool=TranslateTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:46
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
import time
from typing import List

import telebot
from loguru import logger
from telebot import TeleBot

from llmkira.middleware.llm_task import OpenaiMiddleware
from llmkira.receiver import function
from llmkira.receiver.receiver_client import BaseReceiver, BaseSender
from llmkira.receiver.telegram import md2tgmd
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.schema import LlmResult
from llmkira.sdk.schema import Message, File
from llmkira.setting.telegram import BotSetting
from llmkira.task import Task, TaskHeader

__receiver__ = "telegram"

from llmkira.middleware.router.schema import router_set

router_set(role="receiver", name=__receiver__)


class TelegramSender(BaseSender):
    """
    平台路由
    """

    def __init__(self):
        if not BotSetting.available:
            return
        self.bot = TeleBot(token=BotSetting.token)
        from telebot import apihelper

        if BotSetting.proxy_address:
            apihelper.proxy = {"https": BotSetting.proxy_address}
        else:
            apihelper.proxy = None

    async def file_forward(self, receiver: TaskHeader.Location, file_list: List[File]):
        for file_obj in file_list:
            if file_obj.file_url:
                self.bot.send_document(
                    chat_id=receiver.chat_id,
                    document=file_obj.file_url,
                    reply_to_message_id=receiver.message_id,
                    caption=file_obj.file_name,
                )
                continue
            _data: File.Data = await file_obj.raw_file()
            if not _data:
                logger.error(f"file download failed {file_obj.file_id}")
                continue
            if file_obj.file_name.endswith((".jpg", ".png", ".jpeg", ".gif")):
                self.bot.send_photo(
                    chat_id=receiver.chat_id,
                    photo=_data.pair,
                    reply_to_message_id=receiver.message_id,
                    caption=file_obj.caption,
                )
            elif file_obj.file_name.endswith(".webp"):
                self.bot.send_sticker(
                    chat_id=receiver.chat_id,
                    reply_to_message_id=receiver.message_id,
                    sticker=_data.pair,
                )
            elif file_obj.file_name.endswith(".ogg"):
                try:
                    self.bot.send_voice(
                        chat_id=receiver.chat_id,
                        voice=_data.pair,
                        reply_to_message_id=receiver.message_id,
                        caption=file_obj.caption,
                    )
                except telebot.apihelper.ApiTelegramException as e:
                    if "VOICE_MESSAGES_FORBIDDEN" in str(e):
                        self.bot.send_message(
                            chat_id=receiver.chat_id,
                            reply_to_message_id=receiver.message_id,
                            text="I can't send voice message because of your privacy settings.",
                        )
                    else:
                        raise e
            else:
                self.bot.send_document(
                    chat_id=receiver.chat_id,
                    document=_data.pair,
                    reply_to_message_id=receiver.message_id,
                    caption=file_obj.caption,
                )

    async def forward(self, receiver: TaskHeader.Location, message: List[RawMessage]):
        """
        插件专用转发，是Task通用类型
        """
        for item in message:
            await self.file_forward(receiver=receiver, file_list=item.file)
            if item.only_send_file:
                continue
            try:
                self.bot.send_message(
                    chat_id=receiver.chat_id,
                    text=item.text,
                    reply_to_message_id=receiver.message_id,
                    parse_mode="MarkdownV2",
                )
                # TODO Telegram Auto format
            except telebot.apihelper.ApiTelegramException as e:
                time.sleep(1)
                logger.warning(f"telegram send message error, retry\n{e}")
                self.bot.send_message(
                    chat_id=receiver.chat_id,
                    text=md2tgmd.escape(item.text),
                    reply_to_message_id=receiver.message_id,
                )

    async def reply(
        self,
        receiver: TaskHeader.Location,
        message: List[Message],
        reply_to_message: bool = True,
    ):
        """
        模型直转发，Message是Openai的类型
        """
        for item in message:
            raw_message = await self.loop_turn_from_openai(
                platform_name=__receiver__, message=item, locate=receiver
            )

            await self.file_forward(receiver=receiver, file_list=raw_message.file)
            if raw_message.only_send_file:
                continue
            if not raw_message.text:
                continue
            assert raw_message.text, "message content is empty"
            try:
                self.bot.send_message(
                    chat_id=receiver.chat_id,
                    text=raw_message.text,
                    reply_to_message_id=receiver.message_id
                    if reply_to_message
                    else None,
                )
            except telebot.apihelper.ApiTelegramException as e:
                if "message to reply not found" in str(e):
                    self.bot.send_message(
                        chat_id=receiver.chat_id, text=raw_message.text
                    )
                else:
                    raise e
        return logger.trace("reply message")

    async def error(self, receiver: TaskHeader.Location, text):
        self.bot.send_message(
            chat_id=receiver.chat_id, text=text, reply_to_message_id=receiver.message_id
        )

    async def function(
        self,
        receiver: TaskHeader.Location,
        task: TaskHeader,
        llm: OpenaiMiddleware,
        llm_result: LlmResult,
    ):
        task_batch = llm_result.default_message.get_executor_batch()
        verify_map = await self.push_task_create_message(
            task=task, task_batch=task_batch, llm_result=llm_result, receiver=receiver
        )
        new_receiver = task.receiver.model_copy()
        new_receiver.platform = __receiver__
        """更新接收者为当前平台，便于创建的函数消息能返回到正确的客户端"""
        new_meta = task.task_meta.pack_loop(
            plan_chain_pending=task_batch, verify_map=verify_map
        )
        """克隆元数据为当前平台"""

        await Task(queue=function.__receiver__).send_task(
            task=TaskHeader.from_function(
                llm_result=llm_result,
                task_meta=new_meta,
                receiver=new_receiver,
                message=task.message,
            )
        )
        """发送打包后的任务数据"""
        del new_meta
        del task
        """清理内存"""


__sender__ = TelegramSender()


class TelegramReceiver(BaseReceiver):
    """
    receive message from telegram
    """

    async def telegram(self):
        self.set_core(sender=__sender__, task=Task(queue=__receiver__))
        if not BotSetting.available:
            logger.warning("Receiver Runtime:TelegramBot Setting empty")
            return None
        logger.success("Receiver Runtime:TelegramBot start")
        await self.task.consuming_task(self.on_message)



# -*- coding: utf-8 -*-
# @Time    : 2023/10/17 下午9:57
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from enum import Enum
from functools import wraps
from typing import Type, Union, Set, List, Callable, Any

from loguru import logger

from .schema import Builder, Parser, AbstractTransfer

__builder__: Set[Type[Builder]] = set()
__parser__: Set[Type[Parser]] = set()


class Locate(Enum):
    sender = 1
    receiver = 0


def resign_transfer():
    """
    装饰器
    """

    def decorator(func: Union[Builder, Parser, Type[Builder], Type[Parser]]):
        if issubclass(func, Builder):
            logger.success(f"📦 [Plugin Builder transfer] {func.__name__}")
            __builder__.add(func)
        elif issubclass(func, Parser):
            logger.success(f"📦 [Plugin Parser transfer] {func.__name__}")
            __parser__.add(func)
        else:
            raise ValueError(f"Resign Transfer Error for unknown func {type(func)} ")

        @wraps(func)
        async def wrapper(*args, **kwargs):
            # 调用执行函数，中间人
            return func(**kwargs)

        return wrapper

    return decorator


class LoopRunner(object):
    pipe_arg: Any = None

    @staticmethod
    def get_receiver_loop(platform_name: str) -> List[Type[AbstractTransfer]]:
        """
        receiver builder
        message: "RawMessage"
        :return platform
        """
        _loop = []
        for _exec_ram in __builder__:
            _exec = _exec_ram()
            if not _exec.sign:
                logger.error(f"receiver_loop metadata None:sign")
                continue
            if not _exec.sign.platform:
                logger.error(f"receiver_loop metadata None:platform")
                continue
            if _exec.sign.priority is None:
                logger.error(f"receiver_loop metadata None:priority")
                continue
            if _exec.sign.agent != "receiver":
                continue
            if _exec.sign.platform.match(platform_name):
                _loop.append(_exec_ram)
        _loop.sort(key=lambda x: x.sign.priority, reverse=True)
        return _loop

    @staticmethod
    def get_sender_loop(platform_name) -> List[Type[AbstractTransfer]]:
        """
        receiver sender
        message: list, file: List[File]
        :return platform
        """
        _loop = []
        for _exec_ram in __parser__:
            _exec = _exec_ram()
            if not _exec.sign:
                logger.error(f"sender_loop metadata None:sign")
                continue
            if not _exec.sign.platform:
                logger.error(f"sender_loop metadata None:platform")
                continue
            if _exec.sign.priority is None:
                logger.error(f"sender_loop metadata None:priority")
                continue
            if _exec.sign.agent == "sender":
                continue
            if _exec.sign.platform.match(platform_name):
                _loop.append(_exec_ram)
        _loop.sort(key=lambda x: x.sign.priority, reverse=True)
        return _loop

    @property
    def result_pipe_arg(self):
        if not self.pipe_arg:
            raise ValueError(f"pipe_arg is None")
        return self.pipe_arg

    async def exec_loop(self, pipe: List[Type[AbstractTransfer]], pipe_arg: dict, validator: Callable = None):
        """
        exec loop
        """
        self.pipe_arg = pipe_arg
        for loop in pipe:
            try:
                if validator:
                    validator(**self.pipe_arg)
                new_pipe_arg = await loop().pipe(self.pipe_arg)
            except Exception as e:
                # logger.info(f"{loop.__name__} exec_loop error {e}, for sign:{loop.sign}")
                logger.debug(f"{loop.__name__} exec_loop error {e}, for sign:{loop.sign}, pipe_arg:{pipe_arg}")
                # 不更新 pipe_arg
            else:
                # logger.info(f"{loop.__name__} exec_loop success, for sign:{loop.sign}")
                logger.debug(f"{loop.__name__} exec_loop success, for sign:{loop.sign}, new_pipe_arg:{new_pipe_arg}")
                self.pipe_arg = new_pipe_arg


from .default_factory import DefaultMessageBuilder, DefaultMessageParser



# -*- coding: utf-8 -*-
# @Time    : 2023/11/9 下午4:29
# @Author  : sudoskys
# @File    : adapter.py
# @Software: PyCharm
from typing import List, Type, Optional, Literal
from typing import TYPE_CHECKING

from loguru import logger
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from .endpoint.schema import LlmRequest, LlmResult
    # from .endpoint.tokenizer import BaseTokenizer


class SingleModel(BaseModel):
    llm_model: str
    token_limit: int
    func_executor: Literal["function_call", "tool_call", "unsupported"]
    request: Type["LlmRequest"]
    response: Type["LlmResult"]
    schema_type: str
    exception: Optional[str] = Field(None, description="exception")


class ModelMeta(object):
    """
    第三方管理分发模型相关的操作类
    """
    model_list: List[SingleModel] = []

    def add_model(self, models: List[SingleModel]):
        for model in models:
            if not model.exception:
                logger.debug(f"🥐 [Model Available] {model.llm_model}")
            self.model_list.append(model)

    def get_by_model_name(self,
                          *,
                          model_name: str
                          ) -> SingleModel:
        for model in self.model_list:
            if model.llm_model == model_name:
                if model.exception:
                    raise NotImplementedError(
                        f"model {model_name} not implemented"
                        f"exception: {model.exception}"
                    )
                return model
        raise LookupError(
            f"model {model_name} not found! "
            f"please check your model name"
        )

    def get_model_list(self):
        return [model.llm_model for model in self.model_list]

    def get_token_limit(self,
                        *,
                        model_name: str
                        ) -> int:
        return self.get_by_model_name(
            model_name=model_name
        ).token_limit

    def get_schema_model_group(self,
                               *,
                               schema_type: str
                               ) -> List[SingleModel]:
        return [
            model for model in self.model_list
            if model.schema_type == schema_type
        ]


SCHEMA_GROUP = ModelMeta()



# -*- coding: utf-8 -*-
# @Time    : 2023/8/21 上午12:06
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from typing import List, Union, Optional
from urllib.parse import urlparse

from llmkira.sdk.adapter import SCHEMA_GROUP
from llmkira.sdk.func_calling import ToolRegister

from .client import UserCostClient, UserConfigClient, UserCost, UserConfig
from .schema import UserDriverMode
from ...sdk.endpoint import Driver


def is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


class CostControl(object):

    @staticmethod
    async def add_cost(cost: UserCost):
        """
        :param cost: cost
        :return: list
        """
        _user_data = await UserCostClient().insert(data=cost)
        return _user_data

    @staticmethod
    async def get_cost_by_uid(uid: str) -> List[UserCost]:
        return await UserCostClient().read_by_uid(uid=uid)


class UserControl(object):
    @staticmethod
    def get_model():
        return SCHEMA_GROUP.get_model_list()

    @staticmethod
    async def get_driver_config(
            uid: str
    ) -> UserConfig.LlmConfig:
        """
        :param uid: user id
        :return: UserConfig.LlmConfig(Token/Driver)
        """
        _user_data = await UserConfigClient().read_by_uid(uid=uid)
        _user_data = _user_data or UserConfig(uid=uid)
        assert _user_data.llm_driver, "user driver is empty"
        return _user_data.llm_driver

    @staticmethod
    def uid_make(
            platform: str,
            user_id: Union[str, int]
    ):
        """
        :param platform: platform.
        :param user_id: user id.
        :raise AssertionError: user id is not str | platform is empty.
        :return: str
        """
        if isinstance(user_id, int):
            user_id = str(user_id)
        assert isinstance(user_id, str), "user id is not str"
        assert platform, "platform is empty"
        return f"{platform}:{user_id}"

    @staticmethod
    async def clear_endpoint(
            uid: str
    ):
        """
        :param uid: user id
        :return: bool
        """
        _user_data = await UserConfigClient().read_by_uid(uid=uid)
        if not _user_data:
            return True
        _user_data.llm_driver.driver = None
        await UserConfigClient().update(uid=uid, data=_user_data)
        return True

    @staticmethod
    async def set_endpoint(
            uid: str,
            api_key: str,
            endpoint: str = None,
            model: str = None,
            org_id: str = None
    ) -> Driver:
        """
        :param uid: user id
        :param api_key: openai api key
        :param endpoint: openai endpoint
        :param model: openai model
        :param org_id: openai org_id id:
        :raise ValidationError: openai model is not valid
        :return: new_driver
        """
        # assert model in MODEL.__args__, f"openai model is not valid,must be one of {MODEL.__args__}"
        if model not in UserControl.get_model():
            model = UserControl.get_model()[0]
        _user_data = await UserConfigClient().read_by_uid(uid=uid)
        _user_data = _user_data or UserConfig(uid=uid)
        new_driver = Driver(endpoint=endpoint, api_key=api_key, model=model, org_id=org_id)
        _user_data.llm_driver.driver = new_driver
        await UserConfigClient().update(uid=uid, data=_user_data)
        return new_driver

    @staticmethod
    async def block_plugin(
            uid: str,
            plugin_name: str
    ) -> list:
        """
        :param uid: user id
        :param plugin_name: plugin name
        :return: list
        """
        if not (plugin_name in ToolRegister().functions):
            raise ValueError(f"plugin {plugin_name} is not exist :(")
        _user_data = await UserConfigClient().read_by_uid(uid=uid)
        _user_data = _user_data or UserConfig(uid=uid)
        _user_data.plugin_subs.block(plugin_name=plugin_name)
        await UserConfigClient().update(uid=uid, data=_user_data)
        return _user_data.plugin_subs.block_list

    @staticmethod
    async def unblock_plugin(
            uid: str,
            plugin_name: str
    ) -> list:
        """
        :param uid: user id
        :param plugin_name: plugin name
        :return: list
        """
        _user_data = UserConfigClient().read_by_uid(uid=uid)
        _user_data = _user_data or UserConfig(uid=uid)
        _user_data.plugin_subs.unblock(plugin_name=plugin_name)
        await UserConfigClient().update(uid=uid, data=_user_data)
        return _user_data.plugin_subs.block_list

    @staticmethod
    async def set_token(
            uid: str,
            token: Optional[str] = None
    ) -> Optional[str]:
        """
        :param uid: user id
        :param token: bind token
        :raise ValidationError: openai model is not valid
        :return: new_driver
        """
        # assert model in MODEL.__args__, f"openai model is not valid,must be one of {MODEL.__args__}"
        _user_data = await UserConfigClient().read_by_uid(uid=uid)
        _user_data = _user_data or UserConfig(uid=uid)
        _user_data.llm_driver.token = token
        await UserConfigClient().update(uid=uid, data=_user_data)
        return token



# -*- coding: utf-8 -*-
# @Time    : 2023/10/27 下午4:05
# @Author  : sudoskys
# @File    : default_trigger.py
# @Software: PyCharm
from . import resign_trigger, Trigger


@resign_trigger(Trigger(on_platform="telegram", action="deny", priority=0))
async def on_chat_message(message: str, uid: str, **kwargs):
    """
    :param message: RawMessage
    :return:
    """
    if "<deny>" in message:
        return True



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:18
# @Author  : sudoskys
# @File    : app.py
# @Software: PyCharm

from dotenv import load_dotenv
from loguru import logger

load_dotenv()
__area__ = "sender"


def run():
    import asyncio

    from llmkira import load_plugins
    from llmkira.sdk import load_from_entrypoint, get_entrypoint_plugins
    from llmkira.setting import StartSetting
    from llmkira.sdk.cache import global_cache_runtime, global_mongodb_runtime
    global_cache_runtime.init_cache(verbose=True)
    global_mongodb_runtime.init_mongodb(verbose=True)
    from .rss import RssAppRunner

    start_setting = StartSetting.from_subdir()
    wait_list = [RssAppRunner().run(interval=60 * 60 * 1)]
    if start_setting.telegram:
        from .telegram import TelegramBotRunner
        wait_list.append(TelegramBotRunner().run())
    if start_setting.discord:
        from .discord import DiscordBotRunner
        wait_list.append(DiscordBotRunner().run())
    if start_setting.kook:
        from .kook import KookBotRunner
        wait_list.append(KookBotRunner().run())
    if start_setting.slack:
        from .slack import SlackBotRunner
        wait_list.append(SlackBotRunner().run())

    # 初始化插件系统
    load_plugins("llmkira/extra/plugins")
    load_from_entrypoint("llmkira.extra.plugin")

    loaded_message = "\n >>".join(get_entrypoint_plugins())
    logger.success(f"\n===========Third Party Plugins Loaded==========\n >>{loaded_message}")

    async def _main(wait_list_):
        await asyncio.gather(
            *wait_list_
        )

    loop = asyncio.get_event_loop()
    loop.run_until_complete(_main(wait_list_=wait_list))



# -*- coding: utf-8 -*-
# @Time    : 2023/10/11 下午3:34
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from typing import Optional

from loguru import logger

from .schema import Chain
from ...sdk.cache import global_cache_runtime


class AuthReloader(object):
    """
    重放任务
    """

    def __init__(self, uid: str = None):
        self.uid = uid

    @classmethod
    def _prefix(cls, uuid: str) -> str:
        """auth:{auth_schema_version}:uuid"""
        return f"auth:v2:{uuid}"

    @classmethod
    def from_form(cls, platform: str, user_id: str):
        _c = cls()
        _c.uid = f"{platform}:{user_id}"
        return _c

    async def save_auth(self,
                        chain: Chain,
                        timeout: int = 60 * 60 * 24
                        ):
        cache = global_cache_runtime.get_redis()
        logger.debug("Resign Auth Point")
        assert chain.creator_uid == self.uid, "Cant Resign Uid Diff From `self`"
        _cache = await cache.set_data(
            key=self._prefix(uuid=chain.uuid),
            value=chain.model_dump_json(),
            timeout=timeout
        )
        return chain.uuid

    async def read_auth(self, uuid: str) -> Optional[Chain]:
        cache = global_cache_runtime.get_redis()
        _cache = await cache.read_data(key=self._prefix(uuid=uuid))
        if not _cache:
            return None
        logger.debug(f"Get Auth Data {_cache} {_cache}")
        chain = Chain.from_redis(_cache)
        if chain.is_expire:
            logger.debug(f"Auth Expire {chain.uuid}")
            return None
        if chain.creator_uid != self.uid:
            logger.debug(f"Not User {self.uid} Created Auth")
            return None
        return chain


class ChainReloader(object):

    def __init__(self, uid: str):
        self.uid = uid

    def _prefix(self) -> str:
        """chain:{auth_schema_version}:uuid"""
        return f"chain:v2:{self.uid}"

    async def add_task(self, chain: Chain) -> str:
        """
        Add A Chain To Redis List
        :param chain Task
        :return uuid
        """
        cache = global_cache_runtime.get_redis()
        await cache.lpush_data(
            key=self._prefix(),
            value=chain.model_dump_json()
        )
        return chain.uuid

    async def get_task(self) -> Optional[Chain]:
        """
        Get Task From Redis List
        :return Optional[Chain]
        """
        cache = global_cache_runtime.get_redis()
        # FIXME 优化为获取对应空间的数据
        # FIXME signas 应该添加 uuid
        redis_raw = await cache.lpop_data(
            key=self._prefix()
        )
        if not redis_raw:
            return None
        chain = Chain.from_redis(redis_raw)
        if chain.is_expire:
            logger.debug(f"Chain Expire {chain.uuid}")
            return None
        return chain



# -*- coding: utf-8 -*-
# @Time    : 2023/10/20 下午5:53
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from typing import List

from khl import HTTPRequester, Cert, api, MessageTypes
from llmkira.middleware.llm_task import OpenaiMiddleware
from llmkira.receiver import function
from llmkira.receiver.receiver_client import BaseReceiver, BaseSender
from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.schema import LlmResult
from llmkira.sdk.schema import File, Message
from llmkira.setting.kook import BotSetting
from llmkira.task import Task, TaskHeader
from loguru import logger

__receiver__ = "kook"

from llmkira.middleware.router.schema import router_set

# 魔法
import nest_asyncio

nest_asyncio.apply()
# 设置路由系统
router_set(role="receiver", name=__receiver__)


class KookSender(BaseSender):
    """
    平台路由
    """

    def __init__(self):
        if not BotSetting.available:
            return
        self.bot = HTTPRequester(cert=Cert(token=BotSetting.token), ratelimiter=None)

    async def create_asset(self, file: File.Data) -> str:
        return (await self.bot.exec_req(api.Asset.create(file=file.pair)))['url']

    async def send_message(self,
                           channel_id: str,
                           user_id: str,
                           dm: bool,
                           message_type: MessageTypes,
                           content: str,
                           ephemeral: bool = False,
                           reply_message_id: str = None
                           ):
        try:
            if dm:
                message = api.DirectMessage.create(
                    target_id=user_id,
                    type=message_type,
                    content=content,
                    quote=reply_message_id
                )
            else:
                message = api.Message.create(
                    target_id=channel_id,
                    type=message_type,
                    content=content,
                    quote=reply_message_id,
                    temp_target_id=user_id if ephemeral else None,
                )
            _msg = await self.bot.exec_req(message)
        except Exception as e:
            raise ValueError(f"Kook msg send failed,{e}")
        return _msg

    async def file_forward(self, receiver: TaskHeader.Location, file_list: List[File]):
        for file_obj in file_list:
            # URL FIRST
            if file_obj.file_url:
                await self.send_message(
                    channel_id=receiver.thread_id,
                    user_id=receiver.user_id,
                    dm=receiver.thread_id == receiver.chat_id,
                    message_type=MessageTypes.FILE,
                    content=file_obj.file_url
                )
            # DATA
            _data: File.Data = await file_obj.raw_file()
            if not _data:
                logger.error(f"file download failed {file_obj.file_id}")
                continue
            if file_obj.file_name.endswith((".jpg", ".png", ".jpeg", ".gif", ".webp")):
                await self.send_message(
                    channel_id=receiver.thread_id,
                    user_id=receiver.user_id,
                    dm=receiver.thread_id == receiver.chat_id,
                    message_type=MessageTypes.IMG,
                    content=await self.create_asset(file=_data)
                )
            else:
                await self.send_message(
                    channel_id=receiver.thread_id,
                    user_id=receiver.user_id,
                    dm=receiver.thread_id == receiver.chat_id,
                    message_type=MessageTypes.FILE,
                    content=await self.create_asset(file=_data)
                )

    async def forward(self, receiver: TaskHeader.Location, message: List[RawMessage]):
        """
        插件专用转发，是Task通用类型
        """
        for item in message:
            await self.file_forward(
                receiver=receiver,
                file_list=item.file
            )
            if item.only_send_file:
                continue
            await self.send_message(
                channel_id=receiver.thread_id,
                user_id=receiver.user_id,
                dm=receiver.thread_id == receiver.chat_id,
                message_type=MessageTypes.KMD,
                content=item.text
            )

    async def reply(self, receiver: TaskHeader.Location, message: List[Message], reply_to_message: bool = True):
        """
        模型直转发，Message是Openai的类型
        """
        for item in message:
            raw_message = await self.loop_turn_from_openai(platform_name=__receiver__, message=item, locate=receiver)
            await self.file_forward(
                receiver=receiver,
                file_list=raw_message.file
            )
            if raw_message.only_send_file:
                continue
            if not raw_message.text:
                continue
            assert raw_message.text, f"message content is empty"
            await self.send_message(
                channel_id=receiver.thread_id,
                user_id=receiver.user_id,
                dm=receiver.thread_id == receiver.chat_id,
                message_type=MessageTypes.KMD,
                content=raw_message.text
            )
        return logger.trace(f"reply message")

    async def error(self, receiver: TaskHeader.Location, text):
        await self.send_message(
            channel_id=receiver.thread_id,
            user_id=receiver.user_id,
            dm=receiver.thread_id == receiver.chat_id,
            message_type=MessageTypes.TEXT,
            content=text
        )

    async def function(self,
                       receiver: TaskHeader.Location,
                       task: TaskHeader,
                       llm: OpenaiMiddleware,
                       llm_result: LlmResult
                       ):
        task_batch = llm_result.default_message.get_executor_batch()
        verify_map = await self.push_task_create_message(
            task=task,
            task_batch=task_batch,
            llm_result=llm_result,
            receiver=receiver
        )
        new_receiver = task.receiver.model_copy(deep=True)
        new_receiver.platform = __receiver__
        """更新接收者为当前平台，便于创建的函数消息能返回到正确的客户端"""
        new_meta = task.task_meta.pack_loop(
            plan_chain_pending=task_batch,
            verify_map=verify_map
        )
        """克隆元数据为当前平台"""

        await Task(queue=function.__receiver__).send_task(
            task=TaskHeader.from_function(
                llm_result=llm_result,
                task_meta=new_meta,
                receiver=new_receiver,
                message=task.message
            )
        )
        """发送打包后的任务数据"""
        del new_meta
        del task
        """清理内存"""


__sender__ = KookSender()


class KookReceiver(BaseReceiver):
    """
    receive message from telegram
    """

    async def kook(self):
        self.set_core(sender=__sender__, task=Task(queue=__receiver__))
        if not BotSetting.available:
            logger.warning("Receiver Runtime:Kook Setting empty")
            return None
        try:
            logger.success("Receiver Runtime:Kook start")
            await self.task.consuming_task(self.on_message)
        except KeyboardInterrupt:
            logger.warning("Kook Receiver shutdown")
        except Exception as e:
            logger.exception(e)
            raise e



# -*- coding: utf-8 -*-
# @Time    : 2023/8/16 上午11:06
# @Author  : sudoskys
# @File    : components.py
# @Software: PyCharm
import base64
import hashlib
import pickle
import time
from abc import abstractmethod, ABC
from io import BytesIO
from typing import Literal, Optional, List, Type, Union
from typing import TYPE_CHECKING

import aiohttp
import shortuuid
from docstring_parser import parse
from loguru import logger
from pydantic import model_validator, BaseModel, Field, PrivateAttr, ConfigDict

from .cache import global_cache_runtime
from .error import ValidationError, CheckError
from .utils import sync, aiohttp_download_file

if TYPE_CHECKING:
    pass


# ATTENTION:禁止调用上层任何schema包，否则会导致循环引用

def generate_uid():
    return shortuuid.uuid()[0:8].upper()


class File(BaseModel):
    """
    请求体
    可提供标准化的文件定义。
    但是不能直接提供接口标准化，因为文件上传需要特殊处理
    """

    class Data(BaseModel):
        file_name: str
        file_data: bytes = Field(None, description="文件数据")

        @property
        def pair(self):
            return self.file_name, self.file_data

    file_name: str = Field(..., description="文件名，不一定和数据匹配")
    file_id: Optional[str] = Field(None, description="文件的数据库ID，用于索引文件数据，所以没有数据就不要填")
    file_url: Optional[str] = Field(None, description="文件URL，不一定有")
    caption: Optional[str] = Field(default='', description="文件注释，展示给LLM")
    created_by: str = Field(default=None, description="Uploader **UID**")
    created_at: int = Field(default=int(time.time()))
    bytes_length: Optional[int] = Field(default=None, description="File Size")

    @model_validator(mode="after")
    def file_id_validator(self):
        if self.file_id:
            assert len(self.file_id) <= 8, "file_id must be less than 8"
            assert isinstance(self.file_id, str), "file_id must be str"
        if self.created_by:
            assert ":" in self.created_by, "created_by must be uid, include `:` ,like `discord:123456`"
        return self

    def __eq__(self, other):
        if isinstance(other, File):
            return (self.file_id == other.file_id) and (self.file_name == other.file_name)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.file_id) + hash(self.file_name)

    def is_user_upload(self, uid: str):
        """
        判断是否是某个用户上传的
        """
        return self.created_by == uid

    @property
    def file_prompt(self):
        """
        FOR LLM
        """
        _comment = '('
        for key, value in self.model_dump().items():
            if value:
                _comment += f"{key}={value},"
            else:
                if key == "file_id":
                    _comment += f"{key}=MissingAlertNeeded,"
        return f"[Attachment{_comment[:-1]})]"

    @staticmethod
    async def download_file_by_id(
            file_id: str,
    ) -> Optional[Data]:
        cache = global_cache_runtime.get_redis()
        file = await cache.read_data(file_id)
        if not file:
            return None
        file_obj: File.Data = pickle.loads(file)
        return file_obj

    async def raw_file(
            self,
    ) -> Optional[Data]:
        cache = global_cache_runtime.get_redis()
        if not self.file_id:
            return None
        file = await cache.read_data(self.file_id)
        if not file:
            return None
        file_obj: File.Data = pickle.loads(file)
        return file_obj

    @classmethod
    async def upload_file(cls,
                          file_name: str,
                          file_data: Union[bytes, BytesIO],
                          creator_uid: str,
                          caption: str = "",
                          file_id: str = None,
                          size_limit: int = 1024 * 1024 * 10,
                          ):
        """
        上传文件，要求用户使用 UID
        """
        assert isinstance(file_data, bytes) or isinstance(file_data, BytesIO), "file_data must be bytes or BytesIO"
        if file_id is None:
            file_id = str(generate_short_md5(file_data))
        cache = global_cache_runtime.get_redis()
        _byte_length = len(file_data)
        if _byte_length > size_limit:
            raise CheckError(f"file size must be less than {size_limit}")
        await cache.set_data(
            key=file_id,
            value=pickle.dumps(File.Data(file_name=file_name, file_data=file_data)),
            timeout=60 * 60 * 24 * 7
        )
        assert len(file_id) <= 8, "file_id must be less than 8"
        assert isinstance(file_id, str), "file_id must be str"
        return cls(file_id=file_id,
                   file_name=file_name,
                   bytes=_byte_length,
                   created_by=creator_uid,
                   caption=caption,
                   )

    @classmethod
    async def upload_file_only_url(cls,
                                   file_name,
                                   file_url,
                                   created_by: str,
                                   caption: str = "",
                                   ):
        return cls(file_id=None,
                   file_url=file_url,
                   file_name=file_name,
                   created_by=created_by,
                   caption=caption,
                   )

    @classmethod
    async def upload_file_by_download_url(cls,
                                          file_name,
                                          file_url,
                                          created_by: str,
                                          caption: str = "",
                                          size_limit: int = 1024 * 1024 * 10,
                                          headers: dict = None,
                                          session: aiohttp.ClientSession = None,
                                          ):
        cache = global_cache_runtime.get_redis()
        file_data = await aiohttp_download_file(file_url,
                                                session=session,
                                                timeout=20,
                                                size_limit=size_limit,
                                                headers=headers
                                                )
        _byte_length = len(file_data)
        _key = generate_short_md5(file_data)
        await cache.set_data(
            key=_key,
            value=pickle.dumps(File.Data(file_name=file_name, file_data=file_data)),
            timeout=60 * 60 * 24 * 7
        )
        return cls(file_id=_key,
                   file_name=file_name,
                   bytes=_byte_length,
                   created_by=created_by,
                   caption=caption,
                   )


class BaseFunction(BaseModel):
    """
    请求体
    只供 Choice 使用
    """

    class FunctionExtra(BaseModel):
        system_prompt: Optional[str] = Field(None, description="系统提示")

        @classmethod
        def default(cls):
            return cls(system_prompt=None)

    _config: FunctionExtra = FunctionExtra.default()
    name: Optional[str] = Field(None, description="函数名称", pattern=r"^[a-zA-Z0-9_]+$")

    def update_config(self, config: "FunctionExtra") -> "BaseFunction":
        self._config = config
        return self

    def update_system_prompt(self, prompt: str) -> "BaseFunction":
        self._config.system_prompt = prompt
        return self

    @property
    def config(self) -> "FunctionExtra":
        return self._config

    def request_final(self,
                      *,
                      schema_model: str):
        return self.model_copy(deep=True)


class Function(BaseFunction):
    """
    请求体
    供外部模组定义并注册函数
    """

    class Parameters(BaseModel):
        type: str = "object"
        properties: dict = {}
        required: List[str] = Field(default=[], description="必填参数")
        model_config = ConfigDict(extra="ignore")

    name: Optional[str] = Field(None, description="函数名称", pattern=r"^[a-zA-Z0-9_]+$")
    description: Optional[str] = None
    parameters: "Parameters" = Field(default_factory=lambda: Function.Parameters(), description="参数")
    model_config = ConfigDict(extra="ignore")

    def request_final(self,
                      *,
                      schema_model: str):
        """
        标准化
        :param schema_model: 适配的模型
        """
        if schema_model.startswith("gpt-"):
            return self.model_copy(deep=True)
        elif schema_model.startswith("chatglm"):
            return self.model_copy(deep=True)
        else:
            raise CheckError(f"unknown model {schema_model}, cant classify model type")

    def add_property(self,
                     property_name: str,
                     property_type: Literal["string", "integer", "number", "boolean", "object", "array"],
                     property_description: str,
                     enum: Optional[tuple] = None,
                     required: bool = False
                     ):
        """
        加注属性
        """
        self.parameters.properties[property_name] = {}
        self.parameters.properties[property_name]['type'] = property_type
        self.parameters.properties[property_name]['description'] = property_description
        if enum:
            self.parameters.properties[property_name]['enum'] = tuple(enum)
        if required:
            self.parameters.required.append(property_name)

    @classmethod
    def parse_from_pydantic(cls,
                            schema_model: Type[BaseModel],
                            plugin_name: str = None,
                            ):
        """
        解析 pydantic 的 schema
        """
        schema = schema_model.model_json_schema()
        docstring = parse(schema.__doc__ or "")
        parameters = {
            k: v for k, v in schema.items() if k not in ("title", "description")
        }
        for param in docstring.params:
            name = param.arg_name
            description = param.description
            if (name in parameters["properties"]) and description:
                if "description" not in parameters["properties"][name]:
                    parameters["properties"][name]["description"] = description

        parameters["required"] = sorted(
            k for k, v in parameters["properties"].items() if "default" not in v
        )

        if "description" not in schema:
            if docstring.short_description:
                schema["description"] = docstring.short_description
            else:
                schema["description"] = (
                    f"Correctly extracted `{cls.__name__}` with all "
                    f"the required parameters with correct types"
                )
        plugin_name = plugin_name or schema["title"]
        return cls(
            name=plugin_name,
            description=schema["description"],
            parameters=parameters,
        )


class FunctionCallCompletion(BaseModel):
    name: str
    arguments: str


class Tool(BaseModel):
    """
    请求体
    """
    type: str = Field(default="function")
    function: "Function"

    def request_final(
            self,
            *,
            schema_model
    ):
        if schema_model.startswith("gpt-"):
            return self
        elif schema_model.startswith("chatglm"):
            return self
        else:
            return self
            # raise CheckError(f"unknown model {schema_model}, cant classify model type")


class ToolChoice(BaseModel):
    """
    请求体
    """
    type: str = Field(default=None)
    function: "BaseFunction" = Field(default=None)


class ToolCallCompletion(ToolChoice):
    """
    响应
    """
    id: str = Field(default=None)
    type: str = Field(default=None)
    function: "FunctionCallCompletion" = Field(default=None)


class TaskBatch(BaseModel):
    id: str = Field(default=None)
    type: str = Field(default=None)
    function: "FunctionCallCompletion" = Field(default=None)

    @classmethod
    def from_tool_call(cls, tool_call: "ToolCallCompletion"):
        if not tool_call:
            return None
        return cls(
            id=tool_call.id,
            type=tool_call.type,
            function=tool_call.function
        )

    @classmethod
    def from_function_call(cls, function_call: "FunctionCallCompletion"):
        if not function_call:
            return None
        return cls(
            id=None,
            type="function",
            function=function_call
        )

    def get_batch_name(self):
        return self.function.name

    def get_batch_args(self):
        return self.function.arguments

    def get_batch_id(self):
        return self.id


class ContentParts(BaseModel):
    """
    请求体
    """

    class Image(BaseModel):
        url: str
        detail: Optional[str] = None

    type: str
    image_url: Optional[str]
    text: Optional[str]


class Message(BaseModel, ABC):
    """
    请求体
    响应体
    """

    class Meta(BaseModel):
        # message_class: Literal[None, "system", "user", "assistant", "tool", "function"] = Field(default="user")
        index_id: str = Field(default_factory=generate_uid, description="消息ID")
        """消息ID"""

        datatime: str = Field(default=time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()), description="消息时间")
        timestamp: int = Field(default=int(time.time()), description="消息时间戳")
        """消息时间"""

        retrieval: bool = Field(default=False, description="是否可检索")
        """是否折叠(对抗过长消息，为小上下文模型设计)，同时证明我们可以在消息嵌入重整策略"""

        extra: dict = Field(default_factory=dict, description="额外信息")
        """额外信息"""

        files: List["File"] = Field(default_factory=list, description="文件信息")
        """文件信息"""

        @classmethod
        def default(cls, message_class):
            return cls(message_class=message_class)

    _meta: "Meta" = PrivateAttr(default=None)
    """元数据"""
    role: str
    content: Union[str, List[ContentParts], List[dict]]

    def get_meta(self) -> Meta:
        return self._meta

    def update_meta(self, meta: "Meta") -> "Message":
        self._meta = meta
        return self

    @property
    def fold(self) -> "Message":
        return self

    def get_functions(self) -> List["TaskBatch"]:
        raise NotImplementedError

    @abstractmethod
    def request_final(self,
                      schema_model: str,

                      ) -> "Message":
        """
        Openai 请求标准格式最终转换根据 message_class 元信息锁定字段
        :param schema_model: 适配的模型
        """
        raise NotImplementedError


class SystemMessage(Message):
    role: str = Field(default="system")
    content: str
    name: Optional[str] = Field(default=None, description="speaker_name", pattern=r"^[a-zA-Z0-9_]+$")

    def request_final(self,
                      *,
                      schema_model: str,
                      ) -> "Message":
        return self


class UserMessage(Message):
    role: str = Field(default="user")
    content: Union[str, List["ContentParts"], List[dict]]
    name: Optional[str] = Field(default=None, description="speaker_name", pattern=r"^[a-zA-Z0-9_]+$")

    @property
    def fold(self) -> "Message":
        """
        折叠过长特殊类型消息。消息类型为程序自有属性
        插件定义 fold_id 引诱查询
        :return: Message
        """
        metadata_str = (f"""[FoldText](fold_id={self._meta.index_id}"""
                        f"""\ntimestamp={self._meta.datatime}"""
                        f"""\ndescription={self.content[:20] + "..."})"""
                        )
        return self.model_copy(
            update={
                "content": metadata_str
            },
            deep=True
        )

    def request_final(self,
                      *,
                      schema_model: str,
                      ) -> "Message":
        """
        Openai 请求标准格式最终转换根据 message_class 元信息锁定字段
        :param schema_model: 适配的模型
        """
        if "vision" in schema_model:
            if isinstance(self.content, str):
                _new_content: List[ContentParts] = [ContentParts(
                    type="text",
                    text=self.content
                )]
                if self._meta.files:
                    for file in self._meta.files:
                        if file.file_url:
                            _new_content.append(
                                ContentParts(
                                    type="image_url",
                                    image_url=file.file_url
                                )
                            )
                        elif file.file_id:
                            if file.file_name.endswith(("jpg", "png", "jpeg", "gif", "webp", "svg")):
                                base64_image = base64.b64encode(
                                    sync(file.raw_file())
                                ).decode('utf-8')
                                _new_content.append(
                                    ContentParts(
                                        type="image_url",
                                        image_url=f"data:image/png;base64,{base64_image}"
                                    )
                                )
                        else:
                            pass
                self.content = _new_content
        return self


class AssistantMessage(Message):
    role: str = Field(default="assistant")
    content: Union[None, str] = Field(default='', description="assistant content")
    name: Optional[str] = Field(default=None, description="speaker_name", pattern=r"^[a-zA-Z0-9_]+$")
    tool_calls: Optional[List[ToolCallCompletion]] = Field(default=None, description="tool calls")
    """a array of tools, for result"""
    function_call: Optional[FunctionCallCompletion] = Field(default=None, description="Deprecated")
    """Deprecated by openai ,for result"""

    @model_validator(mode="after")
    def deprecate_validator(self):
        if self.tool_calls and self.function_call:
            raise ValidationError("sdk param validator:tool_calls and function_call cannot both be provided")
        if self.function_call:
            logger.warning("sdk param validator:function_call is deprecated")
        if self.content is None:
            self.content = ""
        return self

    def get_executor_batch(self) -> List[TaskBatch]:
        """
        获取插件获取对应的无序函数列表，合并新旧标准
        """
        _calls = []
        if self.function_call:
            _calls.append(TaskBatch.from_function_call(self.function_call))
        _calls.extend([TaskBatch.from_tool_call(tools) for tools in self.tool_calls])
        # 去None
        _calls = [_x for _x in _calls if _x]
        return _calls

    @property
    def sign_function(self) -> bool:
        """
        判断是否有函数需要执行
        """
        if self.function_call:
            return True
        if self.tool_calls:
            for tol in self.tool_calls:
                if tol.function:
                    return True
        return False

    def request_final(self,
                      *,
                      schema_model: str,
                      ) -> "Message":
        return self


class ToolMessage(Message):
    role: str = Field(default="tool")
    content: str
    tool_call_id: str

    def request_final(self,
                      *,
                      schema_model: str,
                      ) -> "Message":
        return self


class FunctionMessage(Message):
    role: str = Field(default="function")
    content: str
    name: str

    @model_validator(mode="after")
    def function_validator(self):
        logger.warning("Function Message is deprecated by openai")
        if self.role == "function" and not self.name:
            raise ValidationError("sdk param validator:name must be specified when role is function")
        return self

    def request_final(self,
                      *,
                      schema_model: str,
                      ) -> "Message":
        """
        Openai 请求标准格式最终转换根据 message_class 元信息锁定字段
        :param schema_model: 适配的模型
        """
        return self


def create_short_task(task_desc, refer, role: str = None):
    """
    创建单任务模板
    """
    if not role:
        role = (
            "[RULE]"
            "Please complete the order according to the task description refer to given information, "
            "if can't complete, please reply 'give up'"
        )
    return [
        SystemMessage(
            role="system",
            content=role,
        ),
        UserMessage(
            role="user",
            content=f"{refer} <hint>{task_desc}<hint>",
            name="task"
        )
    ]


def parse_message_dict(item: dict):
    """
    将 dict 实例化，用错误hook覆盖
    """
    role = item.get("role", None)
    if not role:
        return None
    try:
        if role == "assistant":
            _message = AssistantMessage.model_validate(item)
        elif role == "user":
            _message = UserMessage.model_validate(item)
        elif role == "system":
            _message = SystemMessage.model_validate(item)
        elif role == "tool":
            _message = ToolMessage.model_validate(item)
        elif role == "function":
            _message = FunctionMessage.model_validate(item)
        else:
            raise CheckError(f"unknown message type {role}")
    except Exception as e:
        logger.exception(f"[ParseError]\nparse_message_dict:Unknown message type in redis data:\n{e}")
        return None
    else:
        return _message


def standardise_for_request(
        *,
        message: "Message",
        schema_model: str
) -> "Message":
    """
    标准化转换，供请求使用
    """
    if isinstance(message, dict):
        message = Message.model_validate(message)
    if hasattr(message, "message"):
        return message.request_final(schema_model=schema_model)
    else:
        return message


def generate_short_md5(data: Union[bytes, str, BytesIO],
                       *,
                       length: int = 8,
                       upper: bool = True,
                       ):
    assert 0 < length <= 32, "length must be less than 32"
    # 计算 MD5 哈希值
    md5_hash = hashlib.md5()
    if isinstance(data, bytes) or isinstance(data, BytesIO):
        if isinstance(data, BytesIO):
            data = data.getvalue()
        with BytesIO(data) as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5_hash.update(chunk)
    if isinstance(data, str):
        md5_hash.update(data.encode("utf-8"))
    # 获取哈希值的 16 进制表示
    hex_digest = md5_hash.hexdigest()
    # 生成唯一的短ID
    short_id = str(hex_digest[:length])
    if upper:
        short_id = short_id.upper()
    return short_id



# -*- coding: utf-8 -*-


__package__name__ = "llmkira.extra.plugins.alarm"
__plugin_name__ = "set_alarm_reminder"
__openapi_version__ = "20231111"

from llmkira.sdk.func_calling import verify_openapi_version

verify_openapi_version(__package__name__, __openapi_version__)

from pydantic import field_validator, ConfigDict

from llmkira.sdk.schema import Function

import datetime
import re

from loguru import logger
from pydantic import BaseModel

from llmkira.receiver.aps import SCHEDULER
from llmkira.schema import RawMessage
from llmkira.sdk.func_calling import PluginMetadata, BaseTool
from llmkira.sdk.func_calling.schema import FuncPair
from llmkira.task import Task, TaskHeader

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from llmkira.sdk.schema import TaskBatch

alarm = Function(name=__plugin_name__, description="Set a timed reminder (only for minutes)")
alarm.add_property(
    property_name="delay",
    property_description="The delay time, in minutes",
    property_type="integer",
    required=True
)
alarm.add_property(
    property_name="content",
    property_description="reminder content",
    property_type="string",
    required=True
)


class Alarm(BaseModel):
    delay: int
    content: str
    model_config = ConfigDict(extra="allow")

    @field_validator("delay")
    def delay_validator(cls, v):
        if v < 0:
            raise ValueError("delay must be greater than 0")
        return v


async def send_notify(_platform, _meta, _sender: dict, _receiver: dict, _user, _chat, _content: str):
    await Task(queue=_platform).send_task(
        task=TaskHeader(
            sender=TaskHeader.Location.model_validate(_sender),  # 继承发送者
            receiver=TaskHeader.Location.model_validate(_receiver),  # 因为可能有转发，所以可以单配
            task_meta=TaskHeader.Meta.model_validate(_meta),
            message=[
                RawMessage(
                    user_id=_user,
                    chat_id=_chat,
                    text=_content
                )
            ]
        )
    )


class AlarmTool(BaseTool):
    """
    搜索工具
    """
    silent: bool = False
    function: Function = alarm
    keywords: list = ["闹钟", "提醒", "定时", "到点", '分钟']
    pattern: Optional[re.Pattern] = re.compile(r"(\d+)(分钟|小时|天|周|月|年)后提醒我(.*)")
    require_auth: bool = True

    # env_required: list = ["SCHEDULER", "TIMEZONE"]

    def pre_check(self):
        return True

    def func_message(self, message_text, **kwargs):
        """
        如果合格则返回message，否则返回None，表示不处理
        """
        for i in self.keywords:
            if i in message_text:
                return self.function
        # 正则匹配
        if self.pattern:
            match = self.pattern.match(message_text)
            if match:
                return self.function
        return None

    async def failed(self,
                     task: "TaskHeader", receiver: "TaskHeader.Location",
                     exception,
                     env: dict,
                     arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                     **kwargs
                     ):
        _meta = task.task_meta.reply_notify(
            plugin_name=__plugin_name__,
            callback=[TaskHeader.Meta.Callback.create(
                name=__plugin_name__,
                function_response=f"Timer Run Failed: {exception}",
                tool_call_id=pending_task.get_batch_id()
            )],
            write_back=True,
            release_chain=True
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,
                receiver=receiver,
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        text=f"{__plugin_name__} Run failed {exception}"
                    )
                ]
            )
        )

    async def callback(self,
                       task: "TaskHeader", receiver: "TaskHeader.Location",
                       env: dict,
                       arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                       **kwargs
                       ):
        return None

    async def run(self,
                  task: "TaskHeader", receiver: "TaskHeader.Location",
                  arg: dict, env: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                  ):
        """
        处理message，返回message
        """
        _set = Alarm.model_validate(arg)
        _meta = task.task_meta.reply_message(
            plugin_name=__plugin_name__,
            callback=[
                TaskHeader.Meta.Callback.create(
                    name=__plugin_name__,
                    function_response="Timer Run Success",
                    tool_call_id=pending_task.get_batch_id()
                )
            ]
        )

        logger.debug("Plugin --set_alarm {} minutes".format(_set.delay))
        SCHEDULER.add_job(
            func=send_notify,
            id=receiver.uid,
            trigger="date",
            replace_existing=True,
            misfire_grace_time=1000,
            run_date=datetime.datetime.now() + datetime.timedelta(minutes=_set.delay),
            args=[
                task.receiver.platform,
                _meta.model_dump(),
                task.sender.model_dump(), receiver.model_dump(),
                receiver.user_id, receiver.chat_id,
                _set.content
            ]
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,  # 继承发送者
                receiver=receiver,  # 因为可能有转发，所以可以单配
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        text=f"🍖 The alarm is now set,just wait for {_set.delay} min"
                    )
                ]
            )
        )


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Set a timed reminder (only for minutes)",
    usage="直接说，以分钟为单位，如：10分钟后提醒我吃饭",
    openapi_version=__openapi_version__,
    function={
        FuncPair(function=alarm, tool=AlarmTool)
    },
    homepage="https://github.com/LlmKira"
)



# -*- coding: utf-8 -*-
# @Time    : 2023/11/14 上午11:39
# @Author  : sudoskys
# @File    : rabbitmq.py
# @Software: PyCharm
from dotenv import load_dotenv
from loguru import logger
from pydantic import Field, PrivateAttr, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from llmkira.sdk.utils import sync


class RabbitMQ(BaseSettings):
    """
    代理设置
    """
    amqp_dsn: str = Field("amqp://admin:8a8a8a@localhost:5672", validation_alias='AMQP_DSN')
    _verify_status: bool = PrivateAttr(default=False)
    model_config = SettingsConfigDict(env_file='.env', env_file_encoding='utf-8', extra="ignore")

    @model_validator(mode='after')
    def is_connect(self):
        import aio_pika
        try:
            sync(aio_pika.connect_robust(
                self.amqp_dsn
            ))
        except Exception as e:
            self._verify_status = False
            logger.exception(f'\n⚠️ RabbitMQ DISCONNECT, pls set AMQP_DSN in .env\n--error {e} \n--dsn {self.amqp_dsn}')
            raise e
        else:
            self._verify_status = True
            logger.success(f"🍩 RabbitMQ Connect Success")
            if self.amqp_dsn == "amqp://admin:8a8a8a@localhost:5672":
                logger.warning(
                    f"\n⚠️ You Are Using The Default RabbitMQ Password"
                    f"\nMake Sure You Handle The Port `5672` And Set Firewall Rules"
                )
        return self

    @property
    def available(self):
        return self._verify_status

    @property
    def task_server(self):
        return self.amqp_dsn


load_dotenv()
RabbitMQSetting = RabbitMQ()



# -*- coding: utf-8 -*-
# @Time    : 2023/10/23 下午7:44
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm

####
# 此包包含错误计数器，用于统计错误次数，标记错误次数过多的插件。
# 在构造阶段读取用户数据库，合并至 ignore 中。
# 注意，注意回调的实现。
####
from typing import Dict
from typing import TYPE_CHECKING

import wrapt
from loguru import logger

if TYPE_CHECKING:
    from ...schema import Function

__error_table__: Dict[str, int] = {}


def get_error_plugin(error_times: int = 10) -> list:
    """
    获取错误次数过多的插件
    :param error_times: 错误次数
    :return:
    """
    return [k for k, v in __error_table__.items() if v > error_times]


def recover_error_plugin(function_name: str) -> None:
    """
    恢复错误插件
    :param function_name:
    :return:
    """
    __error_table__[function_name] = 0


def resign_plugin_executor(function: "Function",
                           *,
                           handle_exceptions: tuple = (Exception,),
                           exclude_exceptions: tuple = ()
                           ):
    """
    装饰器，先判断是否排除，再判断是否处理
    :param function: 被装饰的函数
    :param handle_exceptions: 处理的异常，只有在此列表中的异常才会被计数
    :param exclude_exceptions: 排除的异常，不会被计数。不可以是宽泛的异常，如 Exception
    :return: 装饰器
    """
    if not handle_exceptions:
        handle_exceptions = (Exception,)
    if Exception in exclude_exceptions or BaseException in exclude_exceptions:
        raise ValueError("Exception and BaseException cant be exclude")
    logger.success(f"📦 [Plugin exception hook] {function.name}")

    @wrapt.decorator  # 保留被装饰函数的元信息
    def wrapper(wrapped, instance, args, kwargs):
        """
        :param wrapped: 被装饰的函数
        :param instance: https://wrapt.readthedocs.io/en/latest/
        :param args: 被装饰函数的参数
        :param kwargs: 被装饰函数的关键字参数
        :return:
        """
        try:
            res = wrapped(*args, **kwargs)
        except Exception as e:
            if e in exclude_exceptions:
                logger.exception(e)
                return {}
            if e in handle_exceptions:
                __error_table__[function.name] = __error_table__.get(function.name, 0) + 1
                logger.exception(e)
            logger.warning(f"📦 [Plugin Not Handle Exception Hook] {function.name} {e}")
        else:
            return res
        return {}

    return wrapper



# -*- coding: utf-8 -*-
# @Time    : 2023/11/8 下午7:27
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from abc import ABC
from typing import Optional, Any, Union, List
from typing import TYPE_CHECKING

import shortuuid
from pydantic import ConfigDict, BaseModel, Field, PrivateAttr

from .tee import Driver

if TYPE_CHECKING:
    from ..schema import AssistantMessage


class LlmResult(BaseModel, ABC):
    """
    LlmResult
    """

    id: Optional[str] = Field(default=str(shortuuid.uuid()))
    object: str
    created: int
    model: str
    choices: list
    usage: Any = None
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    @property
    def result_type(self):
        return self.object

    def ack(self):
        """
        向上转型
        """
        return self

    @property
    def default_message(self) -> "AssistantMessage":
        return self.choices[0].message


class LlmRequest(BaseModel, ABC):
    """
    LlmRequest
    """
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    _config: Driver = PrivateAttr()
    messages: list
    temperature: Optional[float] = 1
    top_p: Optional[float] = None
    n: Optional[int] = 1
    stop: Optional[Union[str, List[str]]] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    seed: Optional[int] = None

    # 用于调试
    __echo: bool = PrivateAttr(default=False)

    @property
    def config(self):
        return self._config

    @property
    def echo(self):
        return self.__echo

    @property
    def model(self):
        if self.config:
            return self.config.model
        return None

    @classmethod
    def init(cls, driver: Driver, **kwargs):
        """
        Init private args
        :param driver: Driver
        :param kwargs: 配置
        :return: LlmRequest Object
        """
        return cls(
            **kwargs
        ).set_driver(driver=driver)

    def set_driver(self, driver: Driver):
        self._config = driver
        return self

    @property
    def schema_map(self) -> dict:
        return {
            "model": True,
            "messages": True,
            "temperature": True,
            "top_p": True,
            "n": True,
            "stop": True,
            "max_tokens": True,
            "seed": True,
            "presence_penalty": True,
            "frequency_penalty": True,
        }

    def create_params(self):
        _arg = self.model_dump(
            exclude_none=True,
            include=self.schema_map
        )
        assert "messages" in _arg, "messages is required"
        _arg["model"] = self.model
        _arg = {
            k: v
            for k, v in _arg.items()
            if v is not None
        }
        return _arg

    def proxy_address(self):
        return self._config.proxy_address

    async def create(
            self,
            **kwargs
    ):
        raise NotImplementedError


class BaseSchemaType(BaseModel, ABC):
    name: str = "base"



# -*- coding: utf-8 -*-
# @Time    : 2023/10/27 下午2:13
# @Author  : sudoskys
# @File    : provider.py
# @Software: PyCharm
from loguru import logger

from llmkira.extra.user import UserControl, UserConfig, UserDriverMode
from llmkira.middleware.service_provider.schema import ProviderSettingObj
from .service_provider import loaded_provider, PublicProvider

if not loaded_provider:
    raise Exception("⚠️ No Any Driver Provider Loaded, Even Public Provider")


class GetAuthDriver(object):
    def __init__(self, uid: str):
        assert uid, "uid is empty"
        self.uid = uid
        self.user = None

    @classmethod
    def from_platform(cls, platform: str, userid: str):
        return cls(uid=UserControl.uid_make(platform=platform, user_id=userid))

    def provide_type(self):
        if not isinstance(self.user, UserConfig.LlmConfig):
            return None
        return self.user.mode

    async def get(self):
        """
        :raise ProviderException: No Any LLM Config Available
        :return: Driver
        """
        self.user = await UserControl.get_driver_config(uid=self.uid)
        assert isinstance(self.user, UserConfig.LlmConfig), "UserConfig.LlmConfig is empty"
        # 配置了自己的私有例
        if self.user.mode == UserDriverMode.private:
            return self.user.driver
        # Public Provider
        if ProviderSettingObj.is_open_everyone:
            provider = PublicProvider()
            logger.debug(f"🍦 Public Provider --name ({provider.name}) --mode ({self.user.mode}) --uid ({self.uid})")
            if await provider.authenticate(
                    uid=self.uid,
                    token=self.user.token, status=self.user.mode):
                return await provider.request_driver(uid=self.uid, token=self.user.token)
        else:
            # 用户需要特别配置 Token
            provider = loaded_provider()
            if await provider.authenticate(
                    uid=self.uid,
                    token=self.user.token,
                    status=self.user.mode
            ):
                return await provider.request_driver(uid=self.uid, token=self.user.token)
        """
        raise ProviderException(
            f"AuthChanged {self.user.provider} >>change>> {loaded_provider.name.upper()}"
            f"\n🥕 Provider({loaded_provider.name})&Mode({self.user.mode})&UID({self.uid})"
            f"\n🍦 Auth Docs: {loaded_provider().config_docs()}"
        )
        """



# -*- coding: utf-8 -*-
# @Time    : 2023/10/27 下午8:24
# @Author  : sudoskys
# @File    : private.py
# @Software: PyCharm
import time

from loguru import logger
from pydantic import BaseModel, Field

from config import provider_settings
from . import resign_provider
from .schema import BaseProvider, ProviderException
from ...sdk.endpoint import Driver

WHITE_LIST = []
if provider_settings.get("private", default=None) is not None:
    WHITE_LIST = provider_settings.private.get("private_white_list", default=[])
    logger.debug(f"🍦 Private Provider Config Loaded, WHITE_LIST({WHITE_LIST})")


class UserToday(BaseModel):
    count: int = 0
    time: int = Field(default=time.strftime("%Y%m%d", time.localtime()))


@resign_provider()
class PrivateProvider(BaseProvider):
    name = "private"

    def __database_key(self, uid: str):
        return f"driver:{self.name}:{uid}"

    def config_docs(self):
        return "This instance is only available to authorized users :)"

    async def authenticate(self, uid, token, status) -> bool:
        if uid in WHITE_LIST:
            return True
        if not Driver.from_public_env().available:
            raise ProviderException(
                "\nYou are using a public and free instance.\nThe current instance key is not configured.",
                provider=self.name
            )
        raise ProviderException(
            "This is a private instance."
            "\nPlease contact the administrator to apply for a private instance."
            f"\n You id is {uid}",
            provider=self.name
        )

    async def request_driver(self, uid, token) -> Driver:
        return Driver.from_public_env()



# -*- coding: utf-8 -*-

"""
bilibili_api.utils.sync

同步执行异步函数
"""
import asyncio
import hashlib
import tempfile
from bisect import bisect_left
from typing import Coroutine, Dict, List, Optional

import aiohttp
import ffmpeg
import nest_asyncio
import shortuuid
from loguru import logger
from telebot import formatting
from telebot.formatting import escape_markdown

nest_asyncio.apply()


def __ensure_event_loop():
    try:
        asyncio.get_event_loop()
    except Exception:
        asyncio.set_event_loop(asyncio.new_event_loop())


def dict2message(data: Dict[str, str]) -> List[str]:
    _res = []
    for key, value in data.items():
        _res.append("".join([formatting.mitalic(key), escape_markdown("="), f"`{escape_markdown(value)}`"]))
    return _res


def sync(coroutine: Coroutine):
    """
    同步执行异步函数，使用可参考 [同步执行异步代码](https://nemo2011.github.io/bilibili-api/#/sync-executor)

    Args:
        coroutine (Coroutine): 异步函数

    Returns:
        该异步函数的返回值
    """
    __ensure_event_loop()
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coroutine)


def sha1_encrypt(string):
    """
    sha1加密算法
    """

    sha = hashlib.sha1(string.encode('utf-8'))
    encrypts = sha.hexdigest()
    return encrypts[:8]


def generate_uid():
    return shortuuid.uuid()[0:8].upper()


async def aiohttp_download_file(url,
                                session: aiohttp.ClientSession = None,
                                timeout=None,
                                size_limit=None,
                                headers=None,
                                **kwargs):
    if not session:
        session = aiohttp.ClientSession()
    async with session as session:
        async with session.get(url,
                               timeout=timeout,
                               headers=headers, **kwargs
                               ) as response:
            if response.status != 200:
                raise Exception("无法下载文件")

            content_length = response.content_length
            if size_limit and content_length and content_length > size_limit:
                raise Exception("文件大小超过限制")

            contents = await response.read()
            return contents


class Ffmpeg(object):
    @staticmethod
    def convert(*,
                input_c: str = "mp3",
                output_c: str = "ogg",
                stream_data: bytes = None,
                quiet=False
                ) -> Optional[bytes]:
        """
        使用ffmpeg转换音频格式
        :param input_c: 输入音频格式
        :param output_c: 输出音频格式
        :param stream_data: 输入音频流
        :param quiet: 是否静默
        """
        if not input_c.startswith("."):
            input_c = "." + input_c
        if not output_c.startswith("."):
            output_c = "." + output_c
        in_fd, temp_filename = tempfile.mkstemp(suffix=input_c, prefix=None, dir=None, text=False)
        out_fd, out_temp_filename = tempfile.mkstemp(suffix=output_c, prefix=None, dir=None, text=False)
        _bytes = None
        try:
            # 写入文件
            with open(temp_filename, "wb") as f:
                f.write(stream_data)
            stream = ffmpeg.input(filename=temp_filename)
            if output_c == ".ogg":
                stream = ffmpeg.output(stream, filename=out_temp_filename, acodec="libopus")
            else:
                stream = ffmpeg.output(stream, filename=out_temp_filename)
            stream = ffmpeg.overwrite_output(stream)
            _ = ffmpeg.run(stream_spec=stream,
                           quiet=quiet
                           )
            # 读取文件
            import os
            _bytes = os.read(out_fd, os.path.getsize(out_temp_filename))
            assert _bytes, "ffmpeg convert failed"
        except Exception as e:
            logger.error(f"ffmpeg convert failed {e}")
            raise e
        finally:
            import os
            os.close(in_fd)
            os.close(out_fd)
            os.remove(out_temp_filename)
            os.remove(temp_filename)
        return _bytes


def prefix_search(wordlist, prefix):
    """
    在有序列表中二分查找前缀
    :param wordlist: 有序列表
    :param prefix: 前缀
    """
    try:
        index = bisect_left(wordlist, prefix)
        return wordlist[index].startswith(prefix)
    except IndexError:
        return False



# -*- coding: utf-8 -*-
# @Time    : 2023/10/30 下午3:42
# @Author  : sudoskys
# @File    : error.py
# @Software: PyCharm
# 更安全的 format
import random


class ReplyNeededError(Exception):
    """
    Raised a error that need reply
    """

    def __init__(self, message: str = None, *args):
        # 拦截 url 信息
        super().__init__(message, *args)


class MappingDefault(dict):
    def __missing__(self, key):
        return key


# (ﾉ>ω<)ﾉ 贴心的错误提示
REQUEST_ERROR_MESSAGE_TEMPLATE = [
    "Seems like you're having a bit of a problem with {error}",
    "just cant {error} 💍",
    "不想回答你的问题\n\n\n{error}",
    "没听清，刚刚有人在说话\n\n\n{error}",
    "你再问一遍呗。\n\n `{error}`",
    "A man hold a gun to my head, and ask me to say:\n`{error}`",
    "Just look at what you've done? A error!\n `{error}`",
    "Damn, A error hits me!\n`{error}`",
    "(ﾉ>ω<)ﾉ︵ERROR！ \n`{error}`",
    "(・∀・)つ︵ERROR！ \n`{error}`",
    "我不是故意的，但是 ———— {error}",
    "你说的对，但是 {error}",
    "（╯－＿－）╯︵ERROR！ \n`{error}`",
    "(／‵Д′)／~ ︵ ︵ ERROR！ \n`{error}`",
    "Chocolate!( ・∀・)っ■ERROR■ \n`{error}`",
    "（╯°Д°）╯︵/(.□ .|) ERROR！ \n`{error}`",
    "╮(﹀_﹀”)╭ ERROR！ \n`{error}`",
    "（；￣д￣）ERROR！ \n`{error}`",
    "（╯' - ')╯ ERROR！ \n`{error}`",
    "（╯°□°）╯︵( .o.) ERROR！ \n`{error}`",
    "您的网络不给力啊，再试一次吧\n`{error}`",
    "上网小心，别被坏人骗了\n`{error}`",
    "我不知道你在说什么，但是我知道你在说什么\n`{error}`",
    "贴心的我，不会让你看到错误的\n`{error}`",
    "上网小贴士：\n`{error}`",
    "听说偶尔看看错误，可以让你更加的快乐\n`{error}`",
    "你也许在期待我会回答你的问题，但是我不会\n`{error}`",
    "你不会用错命令了吧？\n`{error}`",
    "这么小声，还想当海军部长？\n`{error}`",
    "你说话太抽象了，我听不懂\n`{error}`",
]


def get_request_error_message(error: str):
    _txt: str = random.choice(REQUEST_ERROR_MESSAGE_TEMPLATE)
    return _txt.format_map(MappingDefault(error=error))


# 同样贴心的上传错误提示 (ﾉ>ω<)ﾉ
UPLOAD_ERROR_MESSAGE_TEMPLATE = [
    "I cant upload file {filename} to server {error}",
    "we cant upload {filename}:( , because {error}...",
    "Seems like you're having a bit of a problem uploading {filename}\n`{error}`",
    "just cant upload {filename} to server {error} 💍",
    "I dont know why, but I cant upload {filename} to server {error}",
    ":( I dont want to upload {filename} to server\n `{error}`",
    "{error}, {filename} 404",
    "OMG, {filename} ,ERROR UPLOAD, `{error}`",
    "WTF, I CANT UPLOAD {filename} BECAUSE `{error}`",
    "MY PHONE IS BROKEN, I CANT UPLOAD {filename} BECAUSE `{error}`",
    "As a human, I can't upload {filename} for you :( \n `{error}`",
]


def get_upload_error_message(filename: str, error: str):
    _txt: str = random.choice(REQUEST_ERROR_MESSAGE_TEMPLATE)
    return _txt.format_map(MappingDefault(filename=filename, error=error))



# -*- coding: utf-8 -*-
# @Time    : 2023/10/26 下午10:18
# @Author  : sudoskys
# @File    : client.py
# @Software: PyCharm
import time
from abc import ABC
from typing import List, Optional

from pymongo.errors import DuplicateKeyError

from .schema import UserCost, UserConfig

__DB_NAME__ = "llm_bot"

from ...sdk.cache import global_mongodb_runtime


class Client(ABC):
    client = global_mongodb_runtime.get_mongodb().with_database(__DB_NAME__)

    def use_collection(self, collection_name: str):
        return self.client.with_collection(collection_name)


class UserCostClient(Client):
    def __init__(self):
        self.client = self.use_collection("user_cost")

    async def insert(self, data: UserCost):
        await self.client.insert_one(data.model_dump())
        return data

    async def read_by_uid(self, uid: str) -> List[UserCost]:
        _data_list = await self.client.find_many({"uid": uid})
        return [UserCost(**_data) for _data in _data_list]


class UserConfigClient(Client):
    def __init__(self):
        self.client = self.use_collection("user_config")

    async def update(self, uid: str, data: UserConfig, validate: bool = True) -> UserConfig:
        if validate:
            assert data.uid == uid, "update validate:uid 不一致"
        data.__dict__.update({
            "last_use_time": int(time.time()),
            "uid": uid}
        )
        await self.client.collection.create_index(
            [("uid", 1)], unique=True
        )
        try:
            await self.client.insert_one(data.model_dump(mode="json"))
        except DuplicateKeyError:
            await self.client.update_one({"uid": uid}, {"$set": data.model_dump(mode="json")})
        return data

    async def read_by_uid(self, uid: str) -> Optional[UserConfig]:
        _raw_data = await self.client.find_one({"uid": uid})
        if not _raw_data:
            return None
        return UserConfig(**_raw_data)



# -*- coding: utf-8 -*-
# @Time    : 2023/7/10 下午9:40
# @Author  : sudoskys
# @File    : event.py
# @Software: PyCharm

def help_message():
    return """
    /help - HELP YOURSELF
    /chat - Chat with me :)
    /task - Function enable
    /ask - Chat with func_disable, 禁止函数
    /tool - 工具列表
    /clear - 删除自己的记录
    /auth - POWER
    
Private Chat Only:
    /bind - RSS
    /unbind - RSS
    /set_endpoint - <apikey>#<endpoint>
    /clear_endpoint - Clear endpoint and key
    /env - 配置变量,use as shell
    /token - bind your token
    /token_clear - clear your token
    /func_ban - ban function
    /func_unban - unban function

!Please confirm that that bot instance is secure, some plugins may be dangerous on unsafe instance.
"""



# -*- coding: utf-8 -*-
import re
from abc import ABC, abstractmethod
from typing import Literal, Any

from pydantic import BaseModel, Field


class TransferMata(BaseModel):
    """
    注册标头
    """
    platform: re.Pattern
    plugin_name: str
    # 优先级
    priority: int = Field(default=0, ge=-100, le=100)
    # 适用端 (sender/receiver)
    agent: Literal["sender", "receiver", None] = Field(default=None)


class AbstractTransfer(ABC):
    sign: Any

    async def pipe(self, *args, **kwargs) -> Any:
        pass


class Builder(AbstractTransfer):
    """
    Receiver Parser
    消息对象转媒体文件
    """
    sign: TransferMata

    @abstractmethod
    async def pipe(self, arg: dict) -> Any:
        """
        change only_send_file to `True` for send file only
        :return 是否要回复文本,文件列表
        """
        return arg


class Parser(AbstractTransfer):
    """
    Sender Parser
    媒体文件对象转消息对象
    """
    sign: TransferMata

    @abstractmethod
    async def pipe(self, arg: dict) -> Any:
        """
        change file list to RawMessage,pls return few message
        :return 回环消息列表,文件列表
        """
        return arg



# -*- coding: utf-8 -*-
# @Time    : 2023/11/8 下午8:05
# @Author  : sudoskys
# @File    : test_message.py
# @Software: PyCharm
import sys

sys.path.append("..")

from llmkira.sdk.schema import Message, UserMessage


def test_schema():
    try:
        Message(role="user", content="test")
    except TypeError as e:
        pass
    else:
        assert False, "Should Raise Error"
    UserMessage(content="test")



# -*- coding: utf-8 -*-


import cjieba

from ..solo import singleton
from ..summarization import STOPWORDS


def load_stopwords(stopwords_path):
    """
    加载停用词词典
    :param stopwords_path:
    :return:
    """
    with open(stopwords_path, encoding='utf-8') as f:
        return [line.strip() for line in f]


def split_doc(doc):
    separators = ['。', '!', '！', '？', '?']
    for sep in separators:
        doc = doc.replace(sep, sep + '##')
    sentences = doc.split('##')
    return sentences[:-1]


def calculate_sentence_score(sentence, stopwords):
    # jieba_ret = jieba.analyse.extract_tags(sentence, topK=100, withWeight=True)  # , allowPOS=('ns', 'n', 'vn', 'v'))
    jieba_ret = cjieba.extract(sentence, top_k=100, with_weight=True)
    sentence_score = 0
    for word, score in jieba_ret:
        if word not in stopwords:
            sentence_score += score
    return sentence_score


@singleton
class TfidfSummarization:
    def __init__(self, ratio=0.2):
        self.ratio = ratio

    def analysis(self, doc):
        stopwords = load_stopwords(STOPWORDS)

        # 对文本进行分割
        sentences = split_doc(doc)

        # 存放句子顺序
        sentences_order = {sent: idx for idx, sent in enumerate(sentences)}

        # 存放句子得分
        sentences_score = {sent: calculate_sentence_score(sent, stopwords) for sent in sentences}

        # 根据得分，选出较高得分的句子
        sentences_score_order = sorted(sentences_score.items(), key=lambda item: -item[1])[
                                : int(len(sentences) * self.ratio)]

        # 将较高的分的句子按原文本进行排序输出
        selected_sentences = {sent: sentences_order[sent] for sent, score in sentences_score_order}
        summary = ''.join([sent for sent, order in sorted(selected_sentences.items(), key=lambda item: item[1])])

        return summary


if __name__ == '__main__':
    doc = """为科技和产业发展提供更加充分的人才支撑。"""



# -*- coding: utf-8 -*-
# @Time    : 2023/10/24 下午8:35
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm
from typing import List

from ...schema import RawMessage
from ...sdk import get_error_plugin
from ...sdk.func_calling import ToolRegister
from ...sdk.schema import Function
from ...task import TaskHeader


class FunctionReorganize(object):

    def __init__(self, task):
        self.task: TaskHeader = task
        self._functions: List[Function] = []

    async def _ignore(self, error_times) -> None:
        """
        忽略错误插件
        """
        for _plugin in get_error_plugin(error_times=error_times):
            if _plugin in self._functions:
                self._functions.remove(_plugin)

    @staticmethod
    async def unique(functions: List[Function] = None) -> List[Function]:
        _dict = {}
        for function in functions:
            assert isinstance(function, Function), "llm_task.py:function type error,cant unique"
            _dict[function.name] = function
        functions = list(_dict.values())
        return functions

    async def _build(self) -> None:
        """
        重新建造函数表
        """
        functions = []
        if self.task.task_meta.function_enable:
            # 继承函数
            functions = self.task.task_meta.function_list
            if self.task.task_meta.sign_as[0] == 0:
                # 复制救赎
                self.task.task_meta.function_salvation_list = self.task.task_meta.function_list
                functions = []
                # 重整
                for _index, _message in enumerate(self.task.message):
                    _message: RawMessage
                    functions.extend(
                        ToolRegister().filter_pair(
                            key_phrases=_message.text,
                            message_raw=_message,
                            file_list=_message.file,
                            address=(self.task.sender, self.task.receiver),
                        )
                    )
                self.task.task_meta.function_list = functions
        if self.task.task_meta.sign_as[0] == 0:
            # 容错一层旧节点
            functions.extend(self.task.task_meta.function_salvation_list)
        self._functions = await self.unique(functions)

    async def build_arg(self,
                        *,
                        errored_limit: int = 10
                        ) -> List[Function]:
        await self._build()
        self._functions = await self.unique(self._functions)
        # 过滤
        await self._ignore(error_times=errored_limit)
        return self._functions



# -*- coding: utf-8 -*-
# @Time    : 2023/10/28 下午4:19
# @Author  : sudoskys
# @File    : tutorial.py
# @Software: PyCharm
import pathlib
from time import sleep

import elara
from rich.console import Console

elara_client = elara.exe(path=pathlib.Path(__file__).parent / "elara.db", commitdb=True)

tutorial = [
    {
        "cn": "接下来进行一些说明，如果您不想看到这些说明，请使用 --no-tutorial 参数运行入口文件。",
        "en": "Next, some instructions will be given. "
              "If you don’t want to see these instructions, please run the entry file with the --no-tutorial flag."
    },
    {
        "cn": "请您在 .env 文件中填写您的配置信息。"
              "如果您需要配置 provider，请使用 config_dir/provider_settings.toml 文件。"
              "您可以通过 `docker-compose up -f docker-compose.yml` 来测试服务。",
        "en": "Please fill in your configuration information in the .env file. "
              "If you need to configure the provider, please use the config_dir/provider_settings.toml file. "
              "You can test the service by `docker-compose up -f docker-compose.yml`."
    },
    {
        "cn": "数据库 RabbitMQ 的默认端口为 5672，Redis 的默认端口为 6379，MongoDB 的默认端口为 27017。"
              "请您考虑是否需要添加防火墙配置。其中，RabbitMQ 和 MongoDB 均使用默认配置了密码。请查看 .env 文件进行修改。",
        "en": "The default port of the database RabbitMQ is 5672, "
              "the default port of Redis is 6379, and the default port of MongoDB is 27017."
              "Please consider whether to add firewall configuration. "
              "Among them, RabbitMQ and MongoDB use the default configuration password. "
              "Please check the .env file for modification."
    },
    {
        "cn": "请当心您的 .env 文件，其中包含了您的敏感信息。请不要将 .env 文件上传到公共仓库。",
        "en": "Please be careful with your .env file, "
              "which contains your sensitive information. "
              "Please do not upload the .env file to the public repository."
    },
    {
        "cn": "请当心您的 日志文件，其中包含了您的敏感信息。请不要将 日志文件上传到公共仓库。",
        "en": "Please be careful with your log file, which contains your sensitive information. "
              "Please do not upload the log file to the public repository."
    },
    {
        "cn": "如果您在使用过程中遇到了问题，可以在 GitHub 上提出 issue 来完善测试。",
        "en": "If you encounter any problems during use, you can raise an issue on GitHub to improve the test."
    },
]
tutorial_len = len(tutorial)


def show_tutorial(skip_existing: bool = False, pre_step_stop: int = 5, database_key: str = "55123"):
    global tutorial, elara_client, tutorial_len
    lens = elara_client.get(database_key)
    if skip_existing and str(lens) == str(len(tutorial)):
        return None
    # 截取未读的条目
    tutorial = tutorial[lens:] if skip_existing else tutorial
    console = Console()
    print("\n")
    with console.status("[bold green]Working on tasks...[/bold green]") as status:
        index = 0
        while tutorial:
            info = tutorial.pop(0)
            index += 1
            console.print(info["cn"], style="bold cyan")
            console.print(info["en"], style="bold green", end="\n\n")
            for i in range(pre_step_stop):
                status.update(
                    f"[bold green]({index}/{tutorial_len})Remaining {pre_step_stop - i} "
                    f"seconds to next info... [/bold green] "
                )
                sleep(1)

    # 更新进度
    elara_client.set(database_key, tutorial_len)
    sleep(3)



# -*- coding: utf-8 -*-
from llmkira.external.schema import PluginExternal

# ========== CHANGE FOLLOWING BLOCK TO ADD YOUR OWN PLUGIN ========== #
EXPORT = [
    PluginExternal(
        plugin_name="llmbot_plugin_bilisearch",
        plugin_link="https://github.com/LlmKira/llmbot_plugin_bilisearch",
        plugin_desc="通过自然语言调用哔哩哔哩搜索",
        plugin_functions=[
            "search_in_bilibili"
        ],
        org_id=None,
        author_id="sudoskys",
        plugin_install=PluginExternal.Install(
            shell="pip3 install llmbot_plugin_bilisearch",
            pypi="llmbot_plugin_bilisearch",
            github=None,
        )
    )
]
# ========== CHANGE ABOVE BLOCK TO ADD YOUR OWN PLUGIN ==========#



# -*- coding: utf-8 -*-
# @Time    : 2023/10/26 下午11:38
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from abc import ABC, abstractmethod

from pydantic import field_validator, Field

from llmkira.sdk.endpoint import Driver
from pydantic_settings import BaseSettings


class ProviderSetting(BaseSettings):
    provider: str = Field("PUBLIC", validation_alias="SERVICE_PROVIDER")

    @property
    def is_open_everyone(self):
        return self.provider.upper() == "PUBLIC"

    @field_validator("provider")
    def provider_upper(cls, v):
        return v.upper()


ProviderSettingObj = ProviderSetting()


class ProviderException(Exception):

    def __init__(self, message: str, provider: str = None):
        self.message = message
        self.provider = provider

    def __str__(self):
        if self.provider:
            return f"\n🥐 Provider {self.provider} Say:\n{self.message}"
        return f"\n🧊 {self.message}"


class BaseProvider(ABC):
    name: str

    def __init__(self, *args, **kwargs):
        if not self.name:
            raise ProviderException("Provider must have name", provider="BaseProvider")

    @abstractmethod
    def config_docs(self):
        """
        配置文档
        """
        return "Base Provider"

    @abstractmethod
    async def authenticate(self, uid, token, status) -> bool:
        """
        必须提供认证文档
        """
        raise ProviderException("Base Provider auth your token,refer docs", provider=self.name)

    @abstractmethod
    async def request_driver(self, uid, token) -> Driver:
        """
        根据 Token 申请使用 Public Driver
        """
        raise ProviderException("Base Provider cant request driver", provider=self.name)



# -*- coding: utf-8 -*-
# @Time    : 2023/11/8 下午1:41
# @Author  : sudoskys
# @File    : penne.py
# @Software: PyCharm

"""
此文件供上下使用，不可引用外部其他文件
"""
import hashlib
import os
from typing import Optional
from typing import TYPE_CHECKING

from pydantic import field_validator, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from ..error import ValidationError

if TYPE_CHECKING:
    pass


def sha1_encrypt(string):
    """
    sha1加密算法
    """

    sha = hashlib.sha1(string.encode('utf-8'))
    encrypts = sha.hexdigest()
    return encrypts[:8]


class Driver(BaseSettings):
    """
    通用配置
    """
    endpoint: str = Field(default="https://api.openai.com/v1/chat/completions")
    api_key: str = Field(default=None)
    org_id: Optional[str] = Field(default=None)
    model: str = Field(default="gpt-3.5-turbo-0613")
    retrieve_model: str = Field(default="gpt-3.5-turbo-16k")
    proxy_address: Optional[str] = Field(None)

    # TODO:AZURE API VERSION

    @property
    def detail(self):
        """
        脱敏
        """
        api_key = "****" + str(self.api_key)[-4:]
        return (
            f"Endpoint: {self.endpoint}\nApi_key: {api_key}\n"
            f"Org_id: {self.org_id}\nModel: {self.model}\nRetrieve_model: {self.retrieve_model}"
        )

    @property
    def available(self):
        """
        检查可用性
        :return:
        """
        return all([self.endpoint, self.api_key, self.model])

    @classmethod
    def from_public_env(cls):
        openai_api_key = os.getenv("OPENAI_API_KEY", None)
        openai_endpoint = os.getenv("OPENAI_API_ENDPOINT", "https://api.openai.com/v1/chat/completions")
        openai_org_id = os.getenv("OPENAI_API_ORG_ID", None)
        openai_model = os.getenv("OPENAI_API_MODEL", "gpt-3.5-turbo-0613")
        openai_model_retrieve = os.getenv("OPENAI_API_RETRIEVE_MODEL", "gpt-3.5-turbo-16k")
        openai_proxy = os.getenv("OPENAI_API_PROXY", None)
        return cls(
            endpoint=openai_endpoint,
            api_key=openai_api_key,
            org_id=openai_org_id,
            model=openai_model,
            model_retrieve=openai_model_retrieve,
            proxy_address=openai_proxy
        )

    @field_validator("api_key")
    def check_key(cls, v):
        if v:
            if len(str(v)) < 4:
                raise ValidationError("api_key is too short???")
            """
            # 严格检查
            if not len(str(v)) == 51:
                raise ValidationError("api_key must be 51 characters long")
            """
        else:
            raise ValidationError("api_key is required,pls set OPENAI_API_KEY in .env")
        return v

    @property
    def uuid(self):
        """
        取 api key 最后 3 位 和 sha1 加密后的前 8 位
        :return: uuid for token
        """
        _flag = self.api_key[-3:]
        return f"{_flag}:{sha1_encrypt(self.api_key)}"

    model_config = SettingsConfigDict(env_file='.env', env_file_encoding='utf-8', case_sensitive=True,
                                      arbitrary_types_allowed=True, extra="allow")



# -*- coding: utf-8 -*-
# @Time    : 2023/11/9 下午2:05
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
from typing import List, Optional

from pydantic import ConfigDict, BaseModel, Field


class SlackMessageEvent(BaseModel):
    """
    https://api.slack.com/events/message.im
    """

    class SlackFile(BaseModel):
        id: Optional[str] = Field(None, description="id")
        created: int = Field(None, description="created")
        timestamp: int = Field(None, description="timestamp")
        name: Optional[str] = Field(None, description="name")
        title: Optional[str] = Field(None, description="title")
        mimetype: Optional[str] = Field(None, description="mimetype")
        filetype: Optional[str] = Field(None, description="filetype")
        pretty_type: Optional[str] = Field(None, description="pretty_type")
        user: Optional[str] = Field(None, description="user")
        user_team: Optional[str] = Field(None, description="user_team")
        editable: bool = Field(None, description="editable")
        size: int = Field(None, description="size")
        mode: Optional[str] = Field(None, description="mode")
        is_external: bool = Field(None, description="is_external")
        external_type: Optional[str] = Field(None, description="external_type")
        is_public: bool = Field(None, description="is_public")
        public_url_shared: bool = Field(None, description="public_url_shared")
        display_as_bot: bool = Field(None, description="display_as_bot")
        username: Optional[str] = Field(None, description="username")
        url_private: Optional[str] = Field(None, description="url_private")
        url_private_download: Optional[str] = Field(None, description="url_private_download")
        media_display_type: Optional[str] = Field(None, description="media_display_type")
        thumb_64: Optional[str] = Field(None, description="thumb_64")
        thumb_80: Optional[str] = Field(None, description="thumb_80")
        thumb_360: Optional[str] = Field(None, description="thumb_360")
        thumb_360_w: int = Field(None, description="thumb_360_w")
        thumb_360_h: int = Field(None, description="thumb_360_h")
        thumb_160: Optional[str] = Field(None, description="thumb_160")
        original_w: int = Field(None, description="original_w")
        original_h: int = Field(None, description="original_h")
        thumb_tiny: Optional[str] = Field(None, description="thumb_tiny")
        permalink: Optional[str] = Field(None, description="permalink")
        permalink_public: Optional[str] = Field(None, description="permalink_public")
        has_rich_preview: bool = Field(None, description="has_rich_preview")
        file_access: Optional[str] = Field(None, description="file_access")

    client_msg_id: Optional[str] = Field(None, description="client_msg_id")
    type: Optional[str] = Field(None, description="type")
    text: Optional[str] = Field(None, description="text")
    user: Optional[str] = Field(None, description="user")
    ts: Optional[str] = Field(None, description="ts")
    blocks: List[dict] = Field([], description="blocks")
    team: Optional[str] = Field(None, description="team")
    thread_ts: Optional[str] = Field(None, description="thread_ts")
    parent_user_id: Optional[str] = Field(None, description="parent_user_id")
    channel: Optional[str] = Field(None, description="channel")
    event_ts: Optional[str] = Field(None, description="event_ts")
    channel_type: Optional[str] = Field(None, description="channel_type")
    files: List["SlackFile"] = Field(default=[], description="files")
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")



# -*- coding: utf-8 -*-
# @Time    : 2023/9/4 下午10:49
# @Author  : sudoskys
# @File    : __init__.py
# @Software: PyCharm

# 不允许导入此模块
raise ImportError("This module is not allowed to be imported")
# 去噪声算法
from .sublimate import Sublimate



# -*- coding: utf-8 -*-
# @Time    : 2023/3/31 下午9:46
# @Author  : sudoskys
# @File    : mongo.py
# @Software: PyCharm
import asyncio

from loguru import logger
from motor.motor_asyncio import AsyncIOMotorClient


class MongoClientWrapper:
    def __init__(self, url, db_name=None, collection_name=None):
        self.client = AsyncIOMotorClient(url)
        if db_name and collection_name:
            self._set_db_and_collection(db_name, collection_name)

    async def ping(self):
        try:
            await self.client.server_info()  # test connection
            await self.client.list_database_names()  # test authentication
        except Exception as e:
            logger.debug(f'⚠️ Mongodb {e}')
            return False
        return True

    def _copy_with_changes(self, **kwargs):
        new_instance = type(self).__new__(type(self))
        new_instance.__dict__.update(self.__dict__)
        new_instance.__dict__.update(kwargs)
        return new_instance

    def with_database(self, db_name):
        return self._copy_with_changes(db=self.client[db_name])

    def with_collection(self, collection_name):
        return self._copy_with_changes(collection=self.db[collection_name])

    def _set_db_and_collection(self, db_name, collection_name):
        self.db = self.client[db_name]
        self.collection = self.db[collection_name]

    async def insert_one(self, document):
        result = await self.collection.insert_one(document)
        return result

    async def insert_many(self, documents):
        result = await self.collection.insert_many(documents)
        return result

    async def find_one(self, query):
        result = await self.collection.find_one(query)
        return result

    async def find_many(self, query):
        result = await self.collection.find(query).to_list(length=None)
        return result

    async def update_one(self, filter_query, update_query):
        result = await self.collection.update_one(filter_query, update_query)
        return result

    async def delete_one(self, query):
        result = await self.collection.delete_one(query)
        return result


# Test
if __name__ == '__main__':
    async def main():
        # create a new instance of the MongoClientWrapper class
        mongo = MongoClientWrapper('mongodb://localhost:27017').with_database('test').with_collection('test')
        await mongo.ping()

        # set the database and collection to use

        # insert a document
        result = await mongo.insert_one({'name': 'John', 'age': 30})
        print(result.inserted_id)

        # find a document by query
        result = await mongo.find_one({'name': 'John'})
        print(result)

        # update a document
        filter_query = {'name': 'John'}
        update_query = {'$set': {'age': 31}}
        result = await mongo.update_one(filter_query, update_query)
        print(result.modified_count)

        # delete a document
        query = {'name': 'John'}
        result = await mongo.delete_one(query)
        print(result.deleted_count)

        result = await mongo.insert_many(
            [{'i': i} for i in range(2000)])
        print('inserted %d docs' % (len(result.inserted_ids),))


    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



# -*- coding: utf-8 -*-
# @Time    : 2023/10/26 下午4:01
# @Author  : sudoskys
# @File    : __init__.py.py
# @Software: PyCharm


######
# 管道前置触发管理器，注册触发词或禁止触发词
######


import inspect
from functools import wraps
from typing import Literal, List, Callable

from loguru import logger
from pydantic import BaseModel, Field


class Trigger(BaseModel):
    on_func: Callable = None
    on_platform: str
    action: Literal["allow", "deny"] = "allow"
    priority: int = Field(default=0, ge=-100, le=100)
    message: str = Field(default="Trigger deny your message")
    function_enable: bool = Field(default=False)

    def update_func(self, func: callable):
        self.on_func = func
        return self


__trigger_phrases__: List[Trigger] = []


async def get_trigger_loop(platform_name: str, message: str, uid: str = None):
    """
    receiver builder
    message: "RawMessage"
    :return 如果有触发，则返回触发的action，否则返回None 代表没有操作
    """
    sorted(__trigger_phrases__, key=lambda x: x.priority)
    if not message:
        message = ""
    for trigger in __trigger_phrases__:
        if trigger.on_platform == platform_name:
            try:
                if await trigger.on_func(message, uid):
                    return trigger
            except Exception as e:
                logger.error(f"📦 Plugin:trigger error: {e}")
                pass
    return None


def resign_trigger(trigger: Trigger):
    """
    装饰器
    """

    def decorator(func):
        if inspect.iscoroutinefunction(func):
            __trigger_phrases__.append(trigger.update_func(func))
            logger.success(f"📦 [Plugin trigger hook] {trigger.__repr__()}")
        else:
            raise ValueError(f"Resign Trigger Error for func {func} is not async function")

        @wraps(func)
        async def wrapper(*args, **kwargs):
            # 调用执行函数，中间人
            return func(*args, **kwargs)

        return wrapper

    return decorator


from .default_trigger import on_chat_message



# -*- coding: utf-8 -*-
# @Time    : 2023/10/26 下午11:46
# @Author  : sudoskys
# @File    : public.py
# @Software: PyCharm
import time

from loguru import logger
from pydantic import BaseModel, Field

from config import provider_settings
from ...sdk.cache import global_cache_runtime
from . import resign_provider
from .schema import BaseProvider, ProviderException
from ...sdk.endpoint import Driver

QUOTA = 24
WHITE_LIST = []
if provider_settings.get("public", default=None) is not None:
    QUOTA = provider_settings.public.get("public_quota", default=24)
    WHITE_LIST = provider_settings.public.get("public_white_list", default=[])
    logger.debug(f"🍦 Public Provider Config Loaded, QUOTA({QUOTA}) WHITE_LIST({WHITE_LIST})")


class UserToday(BaseModel):
    count: int = 0
    time: int = Field(default=time.strftime("%Y%m%d", time.localtime()))


@resign_provider()
class PublicProvider(BaseProvider):
    name = "public"

    def __database_key(self, uid: str):
        return f"driver:{self.name}:{uid}"

    def config_docs(self):
        return "ConfigDocs:Its a public provider"

    async def authenticate(self, uid, token, status) -> bool:
        _pass = await self.check_times(times=QUOTA, uid=uid)
        if not _pass:
            raise ProviderException(
                "You are using a public instance. You triggered data flood protection today",
                provider=self.name
            )
        if not Driver.from_public_env().available:
            raise ProviderException(
                "You are using a public instance\nBut current instance apikey unavailable",
                provider=self.name
            )
        return True

    async def check_times(self, times: int, uid: str):
        cache = global_cache_runtime.get_redis()
        date = time.strftime("%Y%m%d", time.localtime())
        read = await cache.read_data(self.__database_key(uid=uid))
        if uid in WHITE_LIST:
            return True
        logger.debug(f"🍦 Public Provider Check Times UID({uid}) Read({read})")
        if read:
            _data: UserToday = UserToday.model_validate(read)
            if str(_data.time) != str(date):
                await cache.set_data(self.__database_key(uid=uid), value=UserToday().model_dump())
                return True
            else:
                if _data.count > times:
                    return False
                if _data.count < times:
                    _data.count += 1
                    await cache.set_data(self.__database_key(uid=uid), value=_data.model_dump())
                    return True
        else:
            _data = UserToday()
            await cache.set_data(self.__database_key(uid=uid), value=_data.model_dump())
            return True
        return False

    async def request_driver(self, uid, token) -> Driver:
        return Driver.from_public_env()



# -*- coding: utf-8 -*-
from pydantic import ConfigDict

__package__name__ = "llmkira.extra.plugins.finish"
__plugin_name__ = "finish_conversation"
__openapi_version__ = "20231111"

from llmkira.sdk.func_calling import verify_openapi_version
from llmkira.sdk.schema import Function

verify_openapi_version(__package__name__, __openapi_version__)

from pydantic import BaseModel, Field

from llmkira.schema import RawMessage
from llmkira.sdk.func_calling import BaseTool, PluginMetadata
from llmkira.sdk.func_calling.schema import FuncPair
from llmkira.task import Task, TaskHeader
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from llmkira.sdk.schema import TaskBatch
finish = Function(
    name=__plugin_name__,
    description="The user's question has been fully answered and there is nothing more to add"
)
finish.add_property(
    property_name="comment",
    property_description="end with a question or a comment.(__language: $context)",
    property_type="string"
)


class Finish(BaseModel):
    comment: str = Field(default=":)", description="end with a question or a comment.(__language: $context)")
    model_config = ConfigDict(extra="allow")


class FinishTool(BaseTool):
    """
    搜索工具
    """
    silent: bool = True
    function: Function = finish
    deploy_child: int = 0
    """可部署子任务 0，终结任务链"""

    def pre_check(self):
        return True

    def func_message(self, message_text, **kwargs):
        # 一直返回，永远处理
        return self.function

    async def failed(self,
                     task: "TaskHeader", receiver: "TaskHeader.Location",
                     env: dict,
                     exception,
                     arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                     **kwargs
                     ):
        _meta = task.task_meta.reply_notify(
            plugin_name=__plugin_name__,
            callback=TaskHeader.Meta.Callback.create(
                function_response=f"Run Failed {exception}",
                name=__plugin_name__,
                tool_call_id=pending_task.get_batch_id()
            ),
            write_back=True,
            release_chain=True
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,
                receiver=receiver,
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        text=f"🍖{__plugin_name__} Run Failed：{exception}"
                    )
                ]
            )
        )

    async def callback(self,
                       task: "TaskHeader", receiver: "TaskHeader.Location",
                       env: dict,
                       arg: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                       **kwargs
                       ):
        return True

    async def run(self,
                  task: "TaskHeader", receiver: "TaskHeader.Location",
                  arg: dict, env: dict, pending_task: "TaskBatch", refer_llm_result: dict = None,
                  ):
        """
        处理message，返回message
        """
        _set = Finish.model_validate(arg)
        # META
        _meta = task.task_meta.reply_message(
            plugin_name=__plugin_name__,
            callback=[
                TaskHeader.Meta.Callback.create(
                    name=__plugin_name__,
                    function_response="Finished the conversation",
                    tool_call_id=pending_task.get_batch_id()
                )
            ],
            function_enable=False
        )
        await Task(queue=receiver.platform).send_task(
            task=TaskHeader(
                sender=task.sender,  # 继承发送者
                receiver=receiver,  # 因为可能有转发，所以可以单配
                task_meta=_meta,
                message=[
                    RawMessage(
                        user_id=receiver.user_id,
                        chat_id=receiver.chat_id,
                        text=_set.comment
                    )
                ]
            )
        )


__plugin_meta__ = PluginMetadata(
    name=__plugin_name__,
    description="Finish the conversation",
    usage="all matched",
    openapi_version=__openapi_version__,
    function={
        FuncPair(function=finish, tool=FinishTool)
    },
)



# -*- coding: utf-8 -*-
# @Time    : 2023/11/9 下午3:50
# @Author  : sudoskys
# @File    : tokenizer.py
# @Software: PyCharm
import json
from typing import Union, List, Type

import tiktoken
from loguru import logger
from pydantic import BaseModel


def _pydantic_type(_message):
    if isinstance(_message, BaseModel):
        return _message.model_dump()
    return _message


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..schema import Message


class BaseTokenizer(object):
    def num_tokens_from_messages(self, messages: List[Union[dict, BaseModel, Type[BaseModel]]], model: str) -> int:
        """Return the number of tokens used by a list of messages_box."""
        raise NotImplementedError


class OpenaiTokenizer(BaseTokenizer):
    def num_tokens_from_messages(self,
                                 messages: List[Union[dict, BaseModel, Type[BaseModel]]],
                                 model: str
                                 ) -> int:
        """Return the number of tokens used by a list of messages_box."""
        if hasattr(messages, "request_final"):
            messages: "Message"
            messages = messages.request_final(schema_model=model)
        messages: List[dict] = [_pydantic_type(message) for message in messages]
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            print("Warning: model not found. Using cl100k_base encoding.")
            encoding = tiktoken.get_encoding("cl100k_base")
        if model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
        }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif model == "gpt-3.5-turbo-0301":
            tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif "gpt-3.5-turbo" in model:
            print("Warning: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613.")
            return self.num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
        elif "gpt-4" in model:
            print("Warning: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613.")
            return self.num_tokens_from_messages(messages, model="gpt-4-0613")
        else:
            tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = 1  # if there's a name, the role is omitted
            logger.warning(
                f"""num_tokens_from_messages() is not implemented for model {model}."""
                """:) If you use a no-openai model, """
                """you can [one-api](https://github.com/songquanpeng/one-api) project handle token usage."""
                """or issue https://github.com/LlmKira/Openaibot/issues to request support."""
            )
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                if isinstance(value, dict):
                    value = json.dumps(value, ensure_ascii=False)
                if isinstance(value, list):
                    value = json.dumps(value, ensure_ascii=False)
                if value is None:
                    continue
                _tokens = len(encoding.encode(value))
                num_tokens += _tokens
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3
        # every reply is primed with <|start|>assistant<|message|>
        return num_tokens


def get_tokenizer(model_name: str) -> BaseTokenizer:
    if model_name.startswith("gpt"):
        return OpenaiTokenizer()
    elif model_name.startswith("."):
        raise NotImplementedError(f"sdk.endpoint.get_tokenizer() is not implemented for model {model_name}.")
    else:
        logger.warning(
            f"sdk.endpoint.get_tokenizer() is not implemented for model {model_name}.\n "
            f"Using default tokenizer.[cl100k]"
        )
        return OpenaiTokenizer()



# -*- coding: utf-8 -*-
# @Time    : 2023/10/11 下午3:35
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
import time

import shortuuid
from pydantic import BaseModel, Field, ConfigDict, model_validator, field_validator

from ...task.schema import TaskHeader


class Chain(BaseModel):
    thead_uuid: str = Field(..., description="Thead UUID")
    creator_uid: str = Field(default=str, description="platform:user_id")
    channel: str = Field(..., description="platform channel")
    arg: TaskHeader = Field(..., description="arg")

    uuid: str = Field(
        default=str(shortuuid.uuid()[0:5]).upper(), description="Always Auto Gen"
    )
    expire: int = Field(default=60 * 60 * 24 * 1, description="expire")
    deprecated: bool = Field(default=False, description="deprecated")
    created_times: int = Field(
        default_factory=lambda: int(time.time()), description="created_times"
    )

    model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)

    def set_expire(self, expire: int) -> "Chain":
        self.expire = expire
        return self

    @classmethod
    def create(
        cls, *, creator_uid: str, channel: str, arg: "TaskHeader", expire: int, **kwargs
    ) -> "Chain":
        return cls(
            creator_uid=creator_uid, channel=channel, arg=arg, expire=expire, **kwargs
        )

    @classmethod
    def from_redis(cls, data: dict) -> "Chain":
        """
        从redis中获取,自动检查是否带有 created_times 和 expire
        """
        assert isinstance(data, dict), "From Redis Receive Empty Data"
        if not data.get("created_times"):
            data["deprecated"] = True
        if not data.get("expire"):
            data["deprecated"] = True
        return cls.model_validate(data)

    @property
    def is_expire(self) -> bool:
        return (int(time.time()) - self.created_times > self.expire) or self.deprecated

    @model_validator(mode="before")
    def check_root(cls, values):
        arg = values.get("arg")
        if arg:
            if isinstance(arg, dict):
                values["arg"] = TaskHeader.model_validate(arg)
            if isinstance(arg, TaskHeader):
                values["arg"] = arg.model_copy(deep=True)
        return values

    @field_validator("creator_uid")
    def check_user_id(cls, v):
        if v.count(":") != 1:
            raise ValueError("Chain `creator_uid` Must Be `platform:user_id` Format")
        if not v:
            raise ValueError("Chain `creator_uid` Is Empty")
        return v

    @field_validator("channel")
    def check_address(cls, v):
        if not v:
            raise ValueError("Chain:channel is empty")
        return v



# -*- coding: utf-8 -*-
# @Time    : 2023/11/14 下午3:05
# @Author  : sudoskys
# @File    : test_router.py
# @Software: PyCharm
import sys

sys.path.append("..")
from llmkira.middleware.router import RouterCache, Router  # noqa: E402


def test_schema():
    r1 = RouterCache(router=[])
    print("\n")
    print(r1)
    try:
        RouterCache(router=None)
    except Exception:
        pass
    else:
        assert False, "Should Raise Error"

    r3 = Router(from_="from", to_="to", user_id=2200, rules="rule there", method="chat")
    print("\n")
    print(r3)



# -*- coding: utf-8 -*-
# @Time    : 2023/8/17 下午8:22
# @Author  : sudoskys
# @File    : telegram.py
# @Software: PyCharm
from typing import Optional

from dotenv import load_dotenv
from loguru import logger
from pydantic import Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class TelegramBot(BaseSettings):
    """
    代理设置
    """
    token: Optional[str] = Field(None, validation_alias='TELEGRAM_BOT_TOKEN')
    proxy_address: Optional[str] = Field(None, validation_alias="TELEGRAM_BOT_PROXY_ADDRESS")  # "all://127.0.0.1:7890"
    bot_link: Optional[str] = Field(None, validation_alias='TELEGRAM_BOT_LINK')
    bot_id: Optional[str] = Field(None, validation_alias="TELEGRAM_BOT_ID")
    bot_username: Optional[str] = Field(None, validation_alias="TELEGRAM_BOT_USERNAME")
    model_config = SettingsConfigDict(env_file='.env', env_file_encoding='utf-8', extra="ignore")

    @model_validator(mode='after')
    def bot_validator(self):
        if self.proxy_address:
            logger.success(f"TelegramBot proxy was set to {self.proxy_address}")
        if self.token is None:
            logger.info(f"\n🍀Check:Telegrambot token is empty")
        if self.bot_id is None and self.token:
            try:
                from telebot import TeleBot
                # 创建 Bot
                if self.proxy_address is not None:
                    from telebot import apihelper
                    if "socks5://" in self.proxy_address:
                        self.proxy_address = self.proxy_address.replace("socks5://", "socks5h://")
                    apihelper.proxy = {'https': self.proxy_address}
                _bot = TeleBot(token=self.token).get_me()
                self.bot_id = str(_bot.id)
                self.bot_username = _bot.username
                self.bot_link = f"https://t.me/{self.bot_username}"
            except Exception as e:
                logger.error(f"\n🍀TelegramBot Token Not Set --error {e}")
            else:
                logger.success(
                    f"🍀TelegramBot Init Connection Success --bot_name {self.bot_username} --bot_id {self.bot_id}"
                )
        return self

    @property
    def available(self):
        return self.token is not None


load_dotenv()
BotSetting = TelegramBot()



# -*- coding: utf-8 -*-
# @Time    : 2023/10/14 下午2:16
# @Author  : sudoskys
# @File    : schema.py
# @Software: PyCharm
import time
from typing import TYPE_CHECKING, Dict, Any
from typing import Tuple, List, Union, Optional

import hikari
import khl
import orjson
import shortuuid
from loguru import logger
from pydantic import model_validator, ConfigDict, Field, BaseModel
from telebot import types

from llmkira.schema import RawMessage
from llmkira.sdk.endpoint.schema import LlmResult
from llmkira.sdk.schema import File, Function, ToolMessage, FunctionMessage, TaskBatch

if TYPE_CHECKING:
    from llmkira.sender.slack.schema import SlackMessageEvent


def orjson_dumps(v, *, default):
    # orjson.dumps returns bytes, to match standard json.dumps we need to decode
    return orjson.dumps(v, default=default).decode()


class TaskHeader(BaseModel):
    """
    任务链节点
    """
    model_config = ConfigDict(arbitrary_types_allowed=True)

    class Meta(BaseModel):
        class Callback(BaseModel):
            function_response: str = Field("empty response", description="工具响应内容")
            name: str = Field(None, description="功能名称", pattern=r"^[a-zA-Z0-9_]+$")
            tool_call_id: Optional[str] = Field(None, description="工具调用ID")

            @model_validator(mode="after")
            def check(self):
                """
                检查回写消息
                """
                if not self.tool_call_id and not self.name:
                    raise ValueError("tool_call_id or name must be set")
                return self

            @classmethod
            def create(cls,
                       *,
                       function_response: str,
                       name: str,
                       tool_call_id: Union[str, None] = None
                       ):
                return cls(
                    function_response=function_response,
                    name=name,
                    tool_call_id=tool_call_id,
                )

            def get_tool_message(self) -> Union["ToolMessage"]:
                if self.tool_call_id:
                    return ToolMessage(
                        tool_call_id=self.tool_call_id,
                        content=self.function_response
                    )
                raise ValueError("tool_call_id is empty")

            def get_function_message(self) -> Union["FunctionMessage"]:
                if self.name:
                    return FunctionMessage(
                        name=self.name,
                        content=self.function_response
                    )
                raise ValueError("name is empty")

        """当前链条的层级"""
        sign_as: Tuple[int, str, str, str] = Field(
            default=(0, "root", "default", str(shortuuid.uuid()).upper()[:5]),
            description="签名"
        )

        """函数并行的信息"""
        plan_chain_archive: List[Tuple["TaskBatch", Union[Exception, dict, str]]] = Field(
            default=[],
            description="完成的节点"
        )
        plan_chain_pending: List["TaskBatch"] = Field(default=[], description="待完成的节点")
        plan_chain_length: int = Field(default=0, description="节点长度")
        plan_chain_complete: Optional[bool] = Field(False, description="是否完成此集群")

        """功能状态与缓存"""
        function_enable: bool = Field(False, description="功能开关")
        function_list: List["Function"] = Field([], description="功能列表")
        function_salvation_list: List["Function"] = Field([], description="上回合的功能列表，用于容错")

        """携带插件的写回结果"""
        write_back: bool = Field(False, description="写回消息")
        callback: List["Callback"] = Field(
            default=[],
            description="用于回写，插件返回的消息头，标识 function 的名字"
        )

        """接收器的路由规则"""
        callback_forward: bool = Field(False, description="转发消息")
        callback_forward_reprocess: bool = Field(False, description="转发消息，但是要求再次处理")
        direct_reply: bool = Field(False, description="直接回复,跳过函数处理等")

        release_chain: bool = Field(False, description="是否响应队列中的函数集群拉起请求")
        """部署点的生长规则"""
        resign_next_step: bool = Field(False, description="函数集群是否可以继续拉起其他函数集群")
        run_step_already: int = Field(0, description="函数集群计数器")
        run_step_limit: int = Field(4, description="函数集群计数器上限")

        """函数中枢的依赖变量"""
        verify_uuid: Optional[str] = Field(None, description="认证链的UUID，根据此UUID和 Map 可以确定哪个需要执行")
        verify_map: Dict[str, "TaskBatch"] = Field({},
                                                   description="函数节点的认证信息，经携带认证重发后可通过")
        llm_result: Any = Field(None, description="存储任务的衍生信息源")
        llm_type: Optional[str] = Field(None, description="存储任务的衍生信息源类型")
        extra_args: dict = Field({}, description="提供额外参数")

        model_config = ConfigDict(extra="ignore", arbitrary_types_allowed=True)

        @property
        def task_uuid(self):
            return self.sign_as[3]

        @model_validator(mode="after")
        def check(self):
            if isinstance(self.llm_result, dict):
                if not self.llm_result.get("model"):
                    raise TypeError("Invalid llm_result")

            if not any([self.callback_forward, self.callback_forward_reprocess, self.direct_reply]):
                if self.write_back:
                    logger.warning("you shouldn*t write back without callback_forward or direct_reply")
                    self.write_back = False
            # If it is the root node, it cannot be written back.
            # Because the message posted by the user is always the root node.
            # Writing back will cause duplicate messages.
            # Because the middleware will write the message back
            if self.sign_as[0] == 0 and self.write_back:
                logger.warning("root node shouldn*t write back")
                self.write_back = False
            return self

        @classmethod
        def from_root(cls, release_chain, function_enable, platform: str = "default", **kwargs):
            return cls(
                sign_as=(0, "root", platform, str(shortuuid.uuid()).upper()[:5]),
                release_chain=release_chain,
                function_enable=function_enable,
                **kwargs
            )

        def pack_loop(
                self,
                *,
                plan_chain_pending: List[TaskBatch],
                verify_map: dict,
        ) -> "TaskHeader.Meta":
            """
            打包循环信息
            :return: Meta
            """
            if not verify_map:
                verify_map = self.verify_map
            if not plan_chain_pending:
                raise ValueError("plan_chain_pending can't be empty")
            _new = self.model_copy(deep=True)
            _new.plan_chain_pending = plan_chain_pending
            _new.plan_chain_length = len(plan_chain_pending)
            _new.verify_map = verify_map
            return _new

        async def work_pending_task(self,
                                    verify_uuid: str
                                    ) -> Optional[TaskBatch]:
            """
            获取待完成的节点
            如果有 uuid 则优先在 map 里面获取对应的节点名称，根据名称查找pending.
            执行完成后，将节点移动到 archive
            """
            if not self.plan_chain_pending:
                self.plan_chain_complete = True
                return logger.trace("no pending task")
            if verify_uuid:
                _task_batch = self.verify_map.get(verify_uuid, None)
                if _task_batch:
                    return _task_batch
            return self.plan_chain_pending[0]

        async def complete_task(self,
                                task_batch: TaskBatch,
                                run_result: Union[Exception, dict, str]
                                ) -> "TaskHeader.Meta":
            """
            完成此集群
            :return: Meta
            """
            self.plan_chain_archive.append((task_batch, run_result))
            if task_batch in self.plan_chain_pending:
                self.plan_chain_pending.remove(task_batch)
            if not self.plan_chain_pending:
                self.plan_chain_complete = True
            return self

        def get_verify_uuid(self, task_batch: TaskBatch) -> Optional[str]:
            """
            从 Map 获取验证UUID
            """
            for key, item in self.verify_map.items():
                if item == task_batch:
                    return key
            return None

        def is_complete(self,
                        if_end_at: "TaskBatch" = None,
                        num_end: int = 0
                        ) -> bool:
            """
            完成状态,谨慎使用
            :param if_end_at: 偏移量
            :param num_end: 偏移量, 优先级高于 if_end_at
            :return: Meta
            """
            if self.plan_chain_complete:
                return True
            if not self.plan_chain_pending:
                self.plan_chain_complete = True
                return True
            if len(self.plan_chain_pending) - num_end <= 0:
                return True
            if if_end_at:
                if len(self.plan_chain_pending) == 1:
                    if self.plan_chain_pending[0] == if_end_at:
                        return True
            return False

        def child(self, name) -> "TaskHeader.Meta":
            """
            生成副本，仅仅是子节点，继承父节点的功能
            """
            self.sign_as = (self.sign_as[0] + 1, "child", name, self.sign_as[3])
            self.run_step_already += 1
            return self.model_copy(deep=True)

        def chain(self,
                  name,
                  write_back: bool,
                  release_chain: bool
                  ) -> "TaskHeader.Meta":
            """
            生成副本，重置链条
            """
            self.sign_as = (self.sign_as[0] + 1, "chain", name, self.sign_as[3])
            self.run_step_already += 1
            self.callback_forward = False
            self.callback_forward_reprocess = False
            self.direct_reply = False
            self.write_back = write_back
            self.release_chain = release_chain
            return self.model_copy(deep=True)

        def reply_direct(self,
                         *,
                         chain_name: str,
                         ):
            """
            决定路由规则。只通知，不重组函数和初始化任何东西。
            最稳定的回复方式，不会出现任何问题，但是不会有任何功能。
            回复消息,其他什么都不做！
            """
            _child = self.child(chain_name)
            _child.callback_forward = False
            _child.callback_forward_reprocess = False
            _child.direct_reply = True
            _child.write_back = False
            _child.release_chain = False
            _child.function_enable = False
            return _child

        def reply_notify(self,
                         *,
                         plugin_name: str,
                         callback: List[Callback],
                         release_chain: bool,
                         function_enable: bool = False,
                         write_back: bool = None
                         ):
            """
            决定路由规则
            默认回复消息，其他参数自由
            只通知，不重组函数和初始化任何东西。
            :param plugin_name: 插件名称
            :param callback: 元信息
            :param write_back: 是否写回，写回此通知到消息历史,比如插件运行失败了，要不要写回消息历史让AI看到呢
            :param release_chain: 是否释放任务链，是否释放任务链，比如插件运行失败，错误消息发送的同时需要释放任务链防止挖坟
            :param function_enable: 是否开启功能
            :return: Meta
            """
            assert isinstance(callback, list), "callback must be list"
            _child = self.child(plugin_name)
            _child.callback = callback
            _child.callback_forward = True
            _child.callback_forward_reprocess = False
            _child.direct_reply = False
            _child.write_back = write_back
            _child.release_chain = release_chain
            _child.function_enable = function_enable
            return _child

        def reply_raw(self,
                      *,
                      plugin_name: str,
                      callback: List[Callback],
                      function_enable: bool = True
                      ):
            """
            决定路由规则
            回写消息，释放任务链，要求再次处理，回复消息
            :param plugin_name: 插件名称
            :param callback: 元信息
            :param function_enable: 是否允许机器人用函数再次处理
            """
            assert isinstance(callback, list), "callback must be list"
            _child = self.child(plugin_name)
            _child.callback = callback
            _child.callback_forward = True
            _child.callback_forward_reprocess = True
            _child.direct_reply = False
            _child.write_back = True
            _child.release_chain = True
            _child.function_enable = function_enable
            return _child

        def reply_message(self,
                          *,
                          plugin_name: str,
                          callback: List[Callback],
                          function_enable: bool = True
                          ):
            """
            决定路由规则
            回写消息，释放任务链，回复消息
            :param plugin_name: 插件名称
            :param callback: 元信息
            :param function_enable: 是否允许机器人用函数再次处理
            """
            assert isinstance(callback, list), "callback must be list"
            _child = self.child(plugin_name)
            _child.callback = callback
            _child.callback_forward = True
            _child.callback_forward_reprocess = False
            _child.direct_reply = False
            _child.write_back = True
            _child.release_chain = True
            _child.function_enable = function_enable
            return _child

    class Location(BaseModel):
        """
        Union[str, int]
        here .... channel
        """
        platform: str = Field(None, description="platform")
        user_id: Union[str, int] = Field(None, description="user id")
        chat_id: Union[str, int] = Field(None, description="guild id(channel in dm)/Telegram chat id")
        thread_id: Optional[Union[str, int]] = Field(None, description="channel id/Telegram thread")
        message_id: Optional[Union[str, int]] = Field(None, description="message id")

        @model_validator(mode="after")
        def to_string(self):
            for key in ["user_id", "chat_id", "thread_id", "message_id"]:
                if isinstance(getattr(self, key), int):
                    setattr(self, key, str(getattr(self, key)))
            return self

        @property
        def uid(self):
            return f"{self.platform}:{self.user_id}"

    class Plugin(BaseModel):
        name: str = Field(None, description="插件名称")
        is_run_out: bool = Field(False, description="是否运行完毕")
        token_usage: int = Field(0, description="Token 用量")

    task_meta: Meta = Field(Meta(), description="任务元数据")
    sender: Location = Field(..., description="发信人")
    receiver: Location = Field(..., description="接收人")
    message: List[RawMessage] = Field(None, description="消息内容")

    @classmethod
    def from_telegram(cls,
                      message: Union[types.Message],
                      task_meta: Meta,
                      file: List[File] = None,
                      reply: bool = True,
                      hide_file_info: bool = False,
                      deliver_back_message: List[types.Message] = None,
                      trace_back_message: List[types.Message] = None
                      ):
        """
        从telegram消息中构建任务
        """

        # none -> []
        trace_back_message = [] if not trace_back_message else trace_back_message
        file = [] if not file else file
        deliver_back_message = [] if not deliver_back_message else deliver_back_message

        def _convert(_message: types.Message) -> Optional[RawMessage]:
            """
            消息标准化
            """
            if not _message:
                raise ValueError("Message is empty")
            if isinstance(_message, types.Message):
                user_id = _message.from_user.id
                chat_id = _message.chat.id
                text = _message.text if _message.text else _message.caption
                created_at = _message.date
            else:
                raise ValueError(f"Unknown message type {type(_message)}")
            return RawMessage(
                user_id=user_id,
                chat_id=chat_id,
                text=text if text else "(empty message)",
                created_at=str(created_at)
            )

        deliver_message_list: List[RawMessage] = [_convert(msg) for msg in deliver_back_message]

        # A
        _file_name = []
        for _file in file:
            _file_name.append(_file.file_prompt)
        # 转换为标准消息
        head_message = _convert(message)
        assert head_message, "HeadMessage is empty"
        # 附加文件信息
        head_message.file = file
        # 追加元信息
        if not hide_file_info:
            head_message.text += "\n" + "\n".join(_file_name)

        # 追加回溯消息
        message_list = []
        if trace_back_message:
            for item in trace_back_message:
                if item:
                    message_list.append(_convert(item))
        message_list.extend(deliver_message_list)
        message_list.append(head_message)

        # 去掉 None
        message_list = [item for item in message_list if item]

        return cls(
            task_meta=task_meta,
            sender=cls.Location(
                platform="telegram",
                chat_id=message.chat.id,
                user_id=message.from_user.id,
                # dm=message.chat.type == "private",
                message_id=message.message_id if reply else None
            ),
            receiver=cls.Location(
                platform="telegram",
                chat_id=message.chat.id,
                user_id=message.from_user.id,
                message_id=message.message_id if reply else None
            ),
            message=message_list
        )

    @classmethod
    def from_function(cls,
                      llm_result: LlmResult,
                      task_meta: Meta,
                      receiver: Location,
                      message: List[RawMessage] = None
                      ):
        """
        从 Openai LLM Task中构建任务
        """
        # task_meta = task_meta.child("function") 发送到 function 的消息不需加点，因为此时 接收器算发送者
        task_meta.llm_result = llm_result
        task_meta.llm_type = llm_result.__repr_name__()
        return cls(
            task_meta=task_meta,
            sender=receiver,
            receiver=receiver,
            message=message
        )

    @classmethod
    def from_router(cls, from_, to_, user_id, method, message_text):
        _meta_arg = {}
        if method == "task":
            _meta_arg["function_enable"] = True
        elif method == "push":
            _meta_arg["callback_forward"] = True
        elif method == "chat":
            _meta_arg["function_enable"] = False
        task_meta = cls.Meta(
            **_meta_arg
        )

        return cls(
            task_meta=task_meta,
            sender=cls.Location(
                platform=from_,
                chat_id=user_id,
                user_id=user_id,
                message_id=None
            ),
            receiver=cls.Location(
                platform=to_,
                chat_id=user_id,
                user_id=user_id,
                message_id=None
            ),
            message=[
                RawMessage(
                    user_id=user_id,
                    chat_id=user_id,
                    text=message_text,
                    created_at=str(int(time.time()))
                )
            ]
        )

    @classmethod
    def from_discord_hikari(cls,
                            message: hikari.Message,
                            task_meta: Meta,
                            file: List[File] = None,
                            reply: bool = True,
                            hide_file_info: bool = False,
                            deliver_back_message: List[hikari.Message] = None,
                            trace_back_message: List[hikari.Message] = None
                            ):
        # none -> []
        trace_back_message = [] if not trace_back_message else trace_back_message
        file = [] if not file else file
        deliver_back_message = [] if not deliver_back_message else deliver_back_message

        def _convert(_message: hikari.Message) -> Optional[RawMessage]:
            """
            消息标准化
            """
            if not _message:
                raise ValueError("Message is empty")
            if isinstance(_message, hikari.Message):
                user_id = message.author.id
                chat_id = message.guild_id if message.guild_id else message.channel_id
                thread_id = message.channel_id
                text = _message.content
                created_at = _message.created_at.timestamp()
            else:
                raise ValueError(f"Unknown message type {type(_message)}")
            return RawMessage(
                user_id=user_id,
                chat_id=chat_id,
                thread_id=thread_id,
                text=text if text else "(empty message)",
                created_at=str(created_at)
            )

        deliver_message_list: List[RawMessage] = [_convert(msg) for msg in deliver_back_message]

        # A
        _file_name = []
        for _file in file:
            _file_name.append(_file.file_prompt)
        # 转换为标准消息
        head_message = _convert(message)
        assert head_message, "HeadMessage is empty"
        # 附加文件信息
        head_message.file = file
        # 追加元信息
        if not hide_file_info:
            head_message.text += "\n" + "\n".join(_file_name)

        # 追加回溯消息
        message_list = []
        if trace_back_message:
            for item in trace_back_message:
                if item:
                    message_list.append(_convert(item))
        message_list.extend(deliver_message_list)
        message_list.append(head_message)

        # 去掉 None
        message_list = [item for item in message_list if item]

        return cls(
            task_meta=task_meta,
            sender=cls.Location(
                platform="discord_hikari",
                thread_id=message.channel_id,
                chat_id=message.guild_id if message.guild_id else message.channel_id,
                user_id=message.author.id,
                message_id=message.id if reply else None
            ),
            receiver=cls.Location(
                platform="discord_hikari",
                thread_id=message.channel_id,
                chat_id=message.guild_id if message.guild_id else message.channel_id,
                user_id=message.author.id,
                message_id=message.id if reply else None
            ),
            message=message_list
        )

    @classmethod
    def from_kook(cls,
                  message: khl.Message,
                  deliver_back_message: List[khl.Message],
                  trace_back_message: List[khl.Message],
                  task_meta: Meta,
                  hide_file_info: bool = False,
                  file: List[File] = None,
                  reply: bool = True,
                  ):
        # none -> []
        trace_back_message = [] if not trace_back_message else trace_back_message
        file = [] if not file else file
        deliver_back_message = [] if not deliver_back_message else deliver_back_message

        def _convert(_message: khl.Message) -> Optional[RawMessage]:
            """
            消息标准化
            """
            if not _message:
                raise ValueError("Message is empty")
            if isinstance(_message, khl.Message):
                user_id = message.author_id
                chat_id = message.ctx.guild.id if message.ctx.guild else message.ctx.channel.id
                thread_id = message.ctx.channel.id
                text = _message.content
                created_at = _message.msg_timestamp
            else:
                raise ValueError(f"Unknown message type {type(_message)}")
            return RawMessage(
                user_id=user_id,
                chat_id=chat_id,
                thread_id=thread_id,
                text=text if text else "(empty message)",
                created_at=str(created_at)
            )

        deliver_message_list: List[RawMessage] = [_convert(msg) for msg in deliver_back_message]

        # A
        _file_name = []
        for _file in file:
            _file_name.append(_file.file_prompt)
        # 转换为标准消息
        head_message = _convert(message)
        assert head_message, "HeadMessage is empty"
        # 附加文件信息
        head_message.file = file
        # 追加元信息
        if not hide_file_info:
            head_message.text += "\n" + "\n".join(_file_name)

        # 追加回溯消息
        message_list = []
        if trace_back_message:
            for item in trace_back_message:
                if item:
                    message_list.append(_convert(item))
        message_list.extend(deliver_message_list)
        message_list.append(head_message)

        # 去掉 None
        message_list = [item for item in message_list if item]

        return cls(
            task_meta=task_meta,
            sender=cls.Location(
                platform="kook",
                thread_id=message.ctx.channel.id,
                chat_id=message.ctx.guild.id if message.ctx.guild else message.ctx.channel.id,
                user_id=message.author_id,
                message_id=message.id if reply else None
            ),
            receiver=cls.Location(
                platform="kook",
                thread_id=message.ctx.channel.id,
                chat_id=message.ctx.guild.id if message.ctx.guild else message.ctx.channel.id,
                user_id=message.author_id,
                message_id=message.id if reply else None
            ),
            message=message_list
        )

    @classmethod
    def from_slack(cls,
                   message: "SlackMessageEvent",
                   deliver_back_message,
                   task_meta: Meta,
                   hide_file_info: bool = False,
                   file: List[File] = None,
                   reply: bool = True,
                   ):
        """
        https://api.slack.com/methods
        """
        # none -> []
        deliver_back_message = [] if not deliver_back_message else deliver_back_message

        def _convert(_message: "SlackMessageEvent") -> Optional[RawMessage]:
            """
            消息标准化
            """
            if not _message:
                raise ValueError("Message is empty")
            if _message.__repr_name__() == "SlackMessageEvent":
                user_id = message.user
                chat_id = message.channel
                thread_id = message.channel
                text = _message.text
                created_at = message.event_ts
            else:
                raise ValueError(f"Unknown message type {type(_message)}")
            return RawMessage(
                user_id=user_id,
                chat_id=chat_id,
                thread_id=thread_id,
                text=text if text else "(empty message)",
                created_at=str(created_at)
            )

        deliver_message_list: List[RawMessage] = [_convert(msg) for msg in deliver_back_message]
        # A
        _file_prompt = []
        for _file in file:
            _file_prompt.append(_file.file_prompt)
        # 转换为标准消息
        now_message = _convert(message)
        assert now_message, "HeadMessage is empty"
        # 附加文件信息
        now_message.file = file
        # 追加文件元信息
        if not hide_file_info:
            now_message.text += "\n" + "\n".join(_file_prompt)

        message_list = []
        message_list.extend(deliver_message_list)
        message_list.append(now_message)
        # 去掉 None
        message_list = [item for item in message_list if item]

        return cls(
            task_meta=task_meta,
            sender=cls.Location(
                platform="slack",
                thread_id=message.channel,
                chat_id=message.channel,
                user_id=message.user,
                message_id=message.thread_ts if reply else None
            ),
            receiver=cls.Location(
                platform="slack",
                thread_id=message.channel,
                chat_id=message.channel,
                user_id=message.user,
                message_id=message.thread_ts if reply else None
            ),
            message=message_list
        )
